import numpy as np
cimport numpy as np

from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records
from libc.stdint cimport uint8_t
from libc.stdint cimport uint16_t
from libc.stdint cimport uint32_t
from libc.stdint cimport uint64_t

from roc.rap.tasks.utils import sort_data

__all__ = ["CWF", "set_cwf"]

# Numpy array dtype for TDS LFM CWF L1 survey data
# Name convention is lower case

CWF = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", ("uint32", 2)),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("survey_mode", "uint8"),
    ("bia_status_info", ("uint8", 6)),
    ("sampling_rate", "float32"),
    ("cwf_data_artefacts", ("uint8", 8)),
    ("input_config", ("uint8", 8)),
    ("channel_status_info", ("uint8", 8)),
    ("waveform_data", ("int16", 8)),
]

cpdef set_cwf(l0, task):
    cdef:
        Py_ssize_t packet
        Py_ssize_t size
        Py_ssize_t sample
        Py_ssize_t index
        np.ndarray data
        tuple acquisition_time
        uint8_t synchro_flag
        tuple channel_status
        tuple artefacts
        tuple bias_status
        tuple input_config
        float sampling_rate
        uint16_t snapshot_nr
        uint16_t samp_per_ch
        uint16_t current_snapshot
        uint32_t current_sample
        uint8_t filter_coefs
        uint8_t counter_ch
        uint8_t channel
        uint64_t samp_step
        uint64_t epoch

    # name of the packet
    name = "TM_TDS_SCIENCE_LFM_CWF"
    logger.debug("Get data for " + name)

    # get the LFM CWF packet data
    if name in l0["TM"]:
        tm = l0["TM"][name]["source_data"]
    else:
        # rest of the code inside the with statement will not be called
        return np.empty(
            0,
            dtype=CWF,
        )

    # Sort input packets by ascending acquisition times
    tm, __ = sort_data(tm, (
        tm["PA_TDS_ACQUISITION_TIME"][:,1],
        tm["PA_TDS_ACQUISITION_TIME"][:,0],
                    ))

    # suppose for now that we have the good size for the data
    size = np.sum(tm["PA_TDS_LFR_CWF_SAMPS_PER_CH"])
    logger.debug("Number of samples: {0}".format(size))

    # create the data array with good size
    data = np.empty(size, dtype=CWF)
    fill_records(data)

    # get the number of packets
    size = tm["PA_TDS_ACQUISITION_TIME"].shape[0]
    logger.debug("Number of packets: {0}".format(size))

    # initiate the group number and the sample number
    current_snapshot = 0
    current_sample = 0

    # initialize the position in the data array
    index = 0

    # survey mode to normal
    data["survey_mode"][:] = 0

    # loop over packets
    with nogil:
        for packet in range(size):
            with gil:
                # acquisition time
                acquisition_time = (
                    tm["PA_TDS_ACQUISITION_TIME"][packet, 0],
                    tm["PA_TDS_ACQUISITION_TIME"][packet, 1],
                )

                # the flag for time synchronization
                synchro_flag = tm["PA_TDS_ACQUISITION_TIME"][packet, 2]

                # channel status info
                channel_status = (
                    tm["PA_TDS_LFM_CWF_CH1"][packet],
                    tm["PA_TDS_LFM_CWF_CH2"][packet],
                    tm["PA_TDS_LFM_CWF_CH3"][packet],
                    tm["PA_TDS_LFM_CWF_CH4"][packet],
                    tm["PA_TDS_LFM_CWF_CH5"][packet],
                    tm["PA_TDS_LFM_CWF_CH6"][packet],
                    tm["PA_TDS_LFM_CWF_CH7"][packet],
                    tm["PA_TDS_LFM_CWF_CH8"][packet],
                )

                # artefacts
                artefacts = (
                    tm["PA_TDS_CWF_ART_CH1_OR"][packet],
                    tm["PA_TDS_CWF_ART_CH2_OR"][packet],
                    tm["PA_TDS_CWF_ART_CH3_OR"][packet],
                    tm["PA_TDS_CWF_ART_CH4_OR"][packet],
                    tm["PA_TDS_CWF_ART_CH5_OR"][packet],
                    tm["PA_TDS_CWF_ART_CH6_OR"][packet],
                    tm["PA_TDS_CWF_ART_BUF_OVERFLOW"][packet],
                    tm["PA_TDS_CWF_ART_SCM_OFF"][packet],
                )

                # input config
                input_config = (
                    tm["PA_TDS_LFM_CWF_INPUT_CONF1"][packet],
                    tm["PA_TDS_LFM_CWF_INPUT_CONF2"][packet],
                    tm["PA_TDS_LFM_CWF_INPUT_CONF3"][packet],
                    tm["PA_TDS_LFM_CWF_INPUT_CONF4"][packet],
                    tm["PA_TDS_LFM_CWF_INPUT_CONF5"][packet],
                    tm["PA_TDS_LFM_CWF_INPUT_CONF6"][packet],
                    tm["PA_TDS_CWF_ART_MAG_HEATER"][packet],
                    tm["PA_TDS_CWF_ART_SCM_CALIB"][packet],
                )

                # bias status info
                bias_status =  (
                    tm["PA_BIA_ON_OFF"][packet],
                    tm["PA_BIA_MODE_BIAS3_ENABLED"][packet],
                    tm["PA_BIA_MODE_BIAS2_ENABLED"][packet],
                    tm["PA_BIA_MODE_BIAS1_ENABLED"][packet],
                    tm["PA_BIA_MODE_HV_ENABLED"][packet],
                    tm["PA_BIA_MODE_MUX_SET"][packet],
                )

                # snapshot number
                snapshot_nr = tm["PA_TDS_SNAPSHOT_NR"][packet]
                if snapshot_nr != current_snapshot:
                    current_snapshot = snapshot_nr
                    current_sample = 0

                # Samples per sec.
                # (Note f["PA_TDS_LFM_CWF_SAMP_RATE"] = log2(samp_rate))
                sampling_rate = 2 ** (tm["PA_TDS_LFM_CWF_SAMP_RATE"][packet])

                # Get TT2000 Epoch times
                epoch = Time().obt_to_utc(np.array([acquisition_time]), to_tt2000=True)

                # Sampling rate in nanoseconds
                samp_step = <unsigned long>(1.0e9 / sampling_rate)

                # sample per channel
                samp_per_ch = tm["PA_TDS_LFR_CWF_SAMPS_PER_CH"][packet]

                # loop over samples defined in the variable
                for sample in range(samp_per_ch):
                    # populate the data array
                    data["epoch"][index] = epoch + (sample * samp_step)
                    data["scet"][index] = Time.cuc_to_scet(np.array([acquisition_time]))
                    data["acquisition_time"][index, :] = acquisition_time
                    data["synchro_flag"][index] = synchro_flag

                    # channel status info
                    data["channel_status_info"][index] = channel_status

                    # data artefacts
                    data["cwf_data_artefacts"][index] = artefacts

                    # input config
                    data["input_config"][index] = input_config

                    # informations for bias
                    data["bia_status_info"][index] = bias_status

                    # sampling rate
                    data["sampling_rate"][index] = sampling_rate

                    # snapshot number
                    #data["SNAPSHOT_SEQ_NR"][index] = snapshot_nr

                    # number of the sample
                    #data["SAMP_SEQ_NR"][index] = current_sample
                    current_sample += 1

                    # initialization of the counter of activated channel to be able
                    # to place read data correctly inside the block
                    counter_ch = 0

                    # loop over channels and add data
                    for channel in range(8):
                        if channel_status[channel] == 1:
                            data["waveform_data"][index, channel] = tm[
                                "PA_TDS_LFM_CWF_DATA_BLK"
                            ][
                                packet,
                                sample + counter_ch * samp_per_ch,
                            ]
                            counter_ch += 1

                    # increment the index
                    index += 1

    return data



#!/usr/bin/env python3
"""
Basic Parameters library

Functions to convert BP

Authors : B. Katra (LPP), R. Piberne (LPP)
"""
# -*- coding: utf-8 -*-

import numpy as np

def init_struct(struct):
    """
    Initialize structures for BP conversion

    :param self: Object instance.
    :param struct: python structure used for BP conversion.
    :type self: object
    :type struct: structure
    :return:
    :rtype:
    :Example: init_struct(struct)

    """

    #For floating point data to be recorded on 16-bit words :
    struct.nbitexp = 6                  # number of bits for the exponent
    struct.nbitsig = 16 - struct.nbitexp       # number of bits for the significand
    struct.rangesig_16 = (1 << struct.nbitsig)-1  # == 2^nbitsig - 1
    struct.rangesig_14 = (1 << 8)-1
    struct.expmax = 32 + 5
    struct.expmin = struct.expmax - (1 << struct.nbitexp) + 1

def convToFloat16(struct, rFloat) :
    """
    BP1 conversion from integer to float16

    :param self: Object instance.
    :param struct: python structure used for BP conversion.
    :param rFloat: Unsigned short (np.uint16_t) to convert
    :type self: object
    :type struct: structure
    :return: The corresponding float
    :rtype: np.float64
    :Example: cdfL2['PB'][...] = convToFloat16(struct, cdfL1['PB'][...])

    """


    #print(rFloat, type(rFloat))
    #print(bin(rFloat))

    #We extract A_exponent and A_significant parts
    A_exp = np.uint8((rFloat>>10) & 0x3f)
    A_sig = (rFloat & 0x3ff)

    #print(A_exp, bin(A_exp))
    #print(A_sig, bin(A_sig))


    #We compute values of significand and exponent
    exp = np.int32(A_exp)+np.int32(struct.expmin)
    sig = np.float64((np.int32(A_sig)) / np.float64((np.int32(struct.rangesig_16))) + 1)/2
    sig_square = sig*2.**exp

    return sig_square


def convToFloat8(rUint) :
    """
    BP1 conversion from integer to float8

    :param self: Object instance.
    :param rUint: Unsigned char (np.uint8_t) to convert
    :type self: object
    :type rUint: np.uint8_t
    :return: The corresponding float
    :rtype: np.float32
    :Example: cdfL2['NVEC_V0'][...] = convToFloat8(cdfL1['NVEC_V0'][...])

    """

    return np.float32(np.int32(rUint)/127.5)-1


def convToFloat4(rUint) :
    """
    BP1 conversion from integer to float4

    :param self: Object instance.
    :param rUint: Unsigned char (np.uint8_t) to convert.
    :type self: object
    :type rUint: np.uint8_t
    :return: The corresponding float
    :rtype: np.float32
    :Example: cdfL2['ELLIP'][...] = convToFloat4(cdfL1['ELLIP'][...])

    """

    #rUint = (rUint >> 3) & 0xF
    #(not need the bit right-shift operator here since the 4-bits of ELLIP have been written in the L0 file in the first 4th bits space
    rUint = rUint & 0xF # 0xF = 00001111

    return np.float32(np.int32(rUint)/15.0)


def convToFloat3(rUint) :
    """
    BP1 conversion from integer to float3

    :param self: Object instance.
    :param rUint: Unsigned char (np.uint8_t) to convert.
    :type self: object
    :type rUint: np.uint8_t
    :return: The corresponding float
    :rtype: np.float32
    :Example: cdfL2['DOP'][...] = convToFloat3(cdfL1['DOP'][...])

    """
    rUint = rUint & 0x7
    return np.float32(np.int32(rUint)/7.0)


def convToFloat14(struct, rFloat) :
    """
    BP1 conversion from integer to float14

    :param self: Object instance.
    :param struct: python structure used for BP conversion.
    :param rFloat: Unsigned short (np.uint16_t) to convert
    :type self: object
    :type struct: structure
    :return: The corresponding float
    :rtype: np.float64
    :Example: cdfL2['VPHI'][...] = convToFloat14(struct, cdfL1['VPHI'][...])

    """

    #We extract A_exponent and A_significant parts
    A_exp = np.uint8((rFloat>>8) & 0x3f)
    A_sig = (rFloat & 0xff)

    #We compute values of significand and exponent
    exp = np.int32(A_exp)+np.int32(struct.expmin)
    sig = np.float64((np.int32(A_sig))/np.float64((np.int32(struct.rangesig_14))) + 1)/2
    sig_square = sig*2.**exp
    return sig_square

def Uint16ToUint8(uint16_array, offset, length):
    """
    Extract bits from an input 16-bits unsigned integer array.
    Output array is returned as an numpy.uint8.

    :param uint16_array: Unsigned short (np.uint16_t) to convert
    :param offset: bit offset
    :param length: number of bits to extract, provided as a hexa
    :return: array of numpy.uint8 with bits extracted
    """

    #We extract bit(s) from input
    return (np.uint8)((uint16_array>>offset) & length)

def convToAutoFloat(struct, rFloat):
    """
    BP2 conversion from integer to float for auto correlation products

    :param self: Object instance.
    :param struct: python structure used for BP conversion.
    :param rUint: Unsigned short (np.uint8_t) to convert
    :type self: object
    :type struct: structure
    :return: The corresponding float
    :rtype: np.float16
    :Example: cdfL2['AUTO'][...] = convToAutoFloat(struct, cdfL1['AUTO'][...])

    """

    #We extract A_exponent and A_significant parts
    A_exp = np.uint8((rFloat>>10) & 0x3f)
    A_sig = (rFloat & 0x3ff)

    #We compute values of significand and exponent
    exp = np.int32(A_exp)+np.int32(struct.expmin)
    sig = np.float64((np.int32(A_sig))/np.float64((np.int32(struct.rangesig_16))) + 1)/2

    sig_square = sig*2.**exp

    return sig_square

def convToCrossFloat(rUint) :
    """
    BP2 conversion from integer to float for cross correlation products

    :param self: Object instance.
    :param rUint: Unsigned char (np.uint8_t) to convert.
    :type self: object
    :type rUint: np.uint8_t
    :return: The corresponding float
    :rtype: np.float64
    :Example: cdfL2['CROSS_RE'][...] = convToCrossFloat(cdfL1['CROSS_RE'][...])

    """
    return np.float32(np.int32(rUint)/127.5)-1


def convToUint8(rUint):
    """
    BP1 conversion from integer to unsigned 8 bit integer for nvec_v2 sign bit

    :param rUint: Unsigned char to convert.
    :type rUint: np.uint8_t
    :return: The corresponding uint8
    :rtype: np.uint8

    """

    return np.uint8(rUint & 0x1)

def main():
    pass


if __name__ == "__main__":
    main()


#! /usr/bin/env python3
# -*- coding:Utf8 -*-

import os
import sys

from distutils.command.build_ext import build_ext
from distutils.core import Extension
from distutils.errors import CCompilerError
from distutils.errors import DistutilsExecError
from distutils.errors import DistutilsPlatformError

# Need packet_structure from roc.rpl plugin
from roc.rpl import packet_structure

# import numpy and cython
from Cython.Build import cythonize
import numpy

RPL_PACKET_STRUCTURE_PATH = os.path.dirname(packet_structure.__file__)

# C Extensions
with_extensions = os.getenv("RAP_EXTENSIONS", None)

if with_extensions == "1" or with_extensions is None:
    with_extensions = True

if with_extensions == "0" or hasattr(sys, "pypy_version_info"):
    with_extensions = False

extensions = []
if with_extensions:
    extensions = list()
    extensions += cythonize(Extension(
        name='*',
        sources=['roc/rap/tasks/thr/*.pyx'],
        include_dirs=[numpy.get_include(), RPL_PACKET_STRUCTURE_PATH]))

    extensions += cythonize(Extension(
        name='*',
        sources=['roc/rap/tasks/lfr/*.pyx'],
        include_dirs=[numpy.get_include(), RPL_PACKET_STRUCTURE_PATH]))

    extensions += cythonize(Extension(
        name='*',
        sources=['roc/rap/tasks/tds/*.pyx'],
        include_dirs=[numpy.get_include(), RPL_PACKET_STRUCTURE_PATH]))


class BuildFailed(Exception):

    pass


class ExtBuilder(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except (DistutilsPlatformError, FileNotFoundError):
            print("************************************************************")
            print("Cannot compile RAP Cython module!")
            print("************************************************************")

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError, ValueError):
            print("************************************************************")
            print("Cannot compile Cython module!")
            print("************************************************************")


def build(setup_kwargs):
    """
    This function is mandatory in order to build the extensions.
    """
    setup_kwargs.update(
        {"ext_modules": extensions, "cmdclass": {"build_ext": ExtBuilder}}
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import datetime as dt

from roc.rap.tasks.thr.hfr_time_log import get_hfr_delta_times
from roc.rap.tasks.thr.normal_burst_tnr import decommute_normal as tnr_normal_decom
from roc.rap.tasks.thr.normal_burst_tnr import decommute_burst as tnr_burst_decom
from roc.rap.tasks.thr.normal_burst_hfr import decommute_normal as hfr_normal_decom
from roc.rap.tasks.thr.normal_burst_hfr import decommute_burst as hfr_burst_decom
from roc.rap.tasks.thr.calibration_tnr import decommute as tnr_calibration_decom
from roc.rap.tasks.thr.calibration_hfr import decommute as hfr_calibration_decom
from roc.rap.tasks.thr.science_spectral_power import decommute as spectral_power_decom

__all__ = [
    'extract_tnr_data',
    'extract_hfr_data',
    'extract_spectral_power_data',
]


# nanoseconds since POSIX epoch to J2000
POSIX_TO_J2000 = 946728000000000000


@np.vectorize
def convert_ns_to_datetime(value):
    return dt.datetime.utcfromtimestamp((value + POSIX_TO_J2000) / 1000000000)


def extract_tnr_data(l0, task):
    """
    Get the TNR data in a record array that can be easily exported into
    the CDF format.

    :param l0: h5py.h5 object containing RPW L0 packet data
    :param task: POPPy Task instance
    :return: TNR data array
    """
    # Extract source_data from the L0 file
    modes = tnr_decommutation(l0, task)

    # filter from None
    modes = [x for x in modes if x is not None]

    # check not empty
    if len(modes) == 0:
        return None

    return np.concatenate(modes)


def extract_hfr_data(l0, task):
    """
    Get the HFR data in a record array that can be easily exported into
    the CDF format.

    :param l0: h5py.h5 object containing RPW L0 packet data
    :param task: POPPy Task instance
    :return: HFR data array
    """
    # Extract source_data from the L0 file
    modes = hfr_decommutation(l0, task)

    # filter from None
    modes = [x for x in modes if x is not None]

    # check not empty
    if len(modes) == 0:
        return None

    return np.concatenate(modes)


def tnr_decommutation(l0, task):
    """
    Extract source_data from TNR SCIENCE TM packets.

    :param l0: h5py.h5 object containing RPW L0 packet data
    :param task: POPPy task
    :return: tuple of normal, burst and calibration TNR data
    """
    normal = None
    burst = None
    calibration = None

    # normal mode
    if 'TM_THR_SCIENCE_NORMAL_TNR' in l0['TM']:
        normal = tnr_normal_decom(
            l0['TM']['TM_THR_SCIENCE_NORMAL_TNR']['source_data'],
            task,
        )

    # burst mode
    if 'TM_THR_SCIENCE_BURST_TNR' in l0['TM']:
        burst = tnr_burst_decom(
            l0['TM']['TM_THR_SCIENCE_BURST_TNR']['source_data'],
            task,
        )

    # calibration mode
    if 'TM_THR_SCIENCE_CALIBRATION_TNR' in l0['TM']:
        calibration = tnr_calibration_decom(
            l0['TM']['TM_THR_SCIENCE_CALIBRATION_TNR']['source_data'],
            task,
        )

    return normal, burst, calibration


def hfr_decommutation(f, task):
    """
    Extract source_data from HFR SCIENCE TM packets.
    """
    normal = None
    burst = None
    calibration = None

    # normal mode
    packet_name = 'TM_THR_SCIENCE_NORMAL_HFR'
    if packet_name in f['TM']:
        # First get actual delta times values from pipeline.hfr_time_log table
        acq_time_min = f['TM'][packet_name]['source_data']['PA_THR_ACQUISITION_TIME'][0][:2]
        mode = 0
        delta_times = get_hfr_delta_times(mode, acq_time_min)

        # Then get measurements in packets
        normal = hfr_normal_decom(
            f['TM'][packet_name]['source_data'],
            task,
            f['TM'][packet_name]['data_field_header']['time'],
            delta_times,
            mode,
        )

    # burst mode
    packet_name = 'TM_THR_SCIENCE_BURST_HFR'
    if packet_name in f['TM']:
        # First get actual delta times values from pipeline.hfr_time_log table
        acq_time_min = f['TM'][packet_name]['source_data']['PA_THR_ACQUISITION_TIME'][0][:2]
        mode = 1
        delta_times = get_hfr_delta_times(mode, acq_time_min)

        # Then get measurements in packets
        burst = hfr_burst_decom(
            f['TM'][packet_name]['source_data'],
            task,
            f['TM'][packet_name]['data_field_header']['time'],
            delta_times,
            mode,
        )

    # calibration mode
    packet_name = 'TM_THR_SCIENCE_CALIBRATION_HFR'
    if packet_name in f['TM']:
        calibration = hfr_calibration_decom(
            f['TM'][packet_name]['source_data'],
            task,
        )

    return normal, burst, calibration


def extract_spectral_power_data(f, task):
    """
    Extract source_data from TM_THR_SCIENCE_SPECTRAL_POWER TM packets.
    """
    if 'TM_THR_SCIENCE_SPECTRAL_POWER' in f['TM']:
        return spectral_power_decom(
            f['TM']['TM_THR_SCIENCE_SPECTRAL_POWER']['source_data'],
            task,
        )
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records

from roc.rap.tasks.utils import sort_data


__all__ = ["STAT", "set_stat"]



# Numpy array dtype for TDS STAT L1 survey data
# Name convention is lower case

STAT = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", ("uint32", 2)),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("survey_mode", "uint8"),
    ("bia_status_info", ("uint8", 6)),
    ("channel_status_info", ("uint8", 4)),
    ("sampling_rate", "uint8"),
    ("rpw_status_info", ("uint8", 8)),
    ("input_config", "uint32"),
    ("snapshot_len", "uint8"),
    ("sn_nr_events", "uint8"),
    ("sn_max_e", "uint16"),
    ("sn_med_max_e", "uint16"),
    ("sn_rms_e", "uint16"),
    ("sn_threshold", "uint16"),
    ("du_nr_impact", "uint8"),
    ("du_med_imp", "uint16"),
    ("wa_amp_max", "uint16"),
    ("wa_amp_med", "uint16"),
    ("wa_rms", "uint16"),
    ("wa_nr_events", "uint8"),
    ("wa_med_freq", "uint8"),
]


def set_stat(l0, task):
    """
    Extract TM_TDS_SCIENCE_NORMAL_STAT data

    :param l0: h5py.h5 object containing the RPW TM/TC packets
    :param task: POPPy task
    :return: numpy array with TDS stat data
    """

    # name of the packet
    name = "TM_TDS_SCIENCE_NORMAL_STAT"
    logger.debug("Get data for " + name)

    # get the TDS STAT packet data
    if name in l0["TM"]:
        tm = l0["TM"][name]["source_data"]
    else:
        # rest of the code inside the with statement will not be called
        return np.empty(
            0,
            dtype=STAT,
        )

    # Sort input packets by ascending acquisition times
    tm, __ = sort_data(tm, (
        tm["PA_TDS_ACQUISITION_TIME"][:,1],
        tm["PA_TDS_ACQUISITION_TIME"][:,0],
                    ))

    # get the number of packets
    size = tm["PA_TDS_ACQUISITION_TIME"].shape[0]
    logger.debug("Number of packets: {0}".format(size))

    # create the data array with good size
    data = np.empty(size, dtype=STAT)
    fill_records(data)

    # set the data for stat, directly a copy of the content of parameters
    data["epoch"][:] = Time().obt_to_utc(tm["PA_TDS_ACQUISITION_TIME"][:, :2], to_tt2000=True)
    data["acquisition_time"][:, :] = tm["PA_TDS_ACQUISITION_TIME"][:, :2]
    data["synchro_flag"][:] = tm["PA_TDS_ACQUISITION_TIME"][:, 2]
    data["scet"][:] = Time.cuc_to_scet(data["acquisition_time"])
    data["survey_mode"][:] = 0

    # bias info
    data["bia_status_info"][:, :] = np.vstack(
        (
            tm["PA_BIA_ON_OFF"],
            tm["PA_BIA_MODE_BIAS3_ENABLED"],
            tm["PA_BIA_MODE_BIAS2_ENABLED"],
            tm["PA_BIA_MODE_BIAS1_ENABLED"],
            tm["PA_BIA_MODE_HV_ENABLED"],
            tm["PA_BIA_MODE_MUX_SET"],
        )
    ).T

    # sampling rate
    data["sampling_rate"] = tm["PA_TDS_STAT_SAMP_RATE"]

    # Channel status info
    data["channel_status_info"] = np.vstack(
        (
        tm["PA_TDS_STAT_ADC_CH1"],
        tm["PA_TDS_STAT_ADC_CH2"],
        tm["PA_TDS_STAT_ADC_CH3"],
        tm["PA_TDS_STAT_ADC_CH4"],
        )
    ).T


    # rpw status
    data["rpw_status_info"][:, :] = np.vstack(
        (
            tm["PA_TDS_THR_OFF"],
            tm["PA_TDS_LFR_OFF"],
            tm["PA_TDS_ANT1_OFF"],
            tm["PA_TDS_ANT2_OFF"],
            tm["PA_TDS_ANT3_OFF"],
            tm["PA_TDS_SCM_OFF"],
            tm["PA_TDS_HF_ART_MAG_HEATER"],
            tm["PA_TDS_HF_ART_SCM_CALIB"],
        )
    ).T

    # input config
    data["input_config"] = tm["PA_TDS_SWF_HF_CH_CONF"]

    # stat parameters
    data["snapshot_len"] = tm["PA_TDS_SNAPSHOT_LEN"]
    data["sn_nr_events"] = tm["PA_TDS_SC_STAT_SN_NR_EVENTS"]
    data["sn_max_e"] = tm["PA_TDS_SC_STAT_SN_MAX_E"]
    data["sn_med_max_e"] = tm["PA_TDS_SC_STAT_SN_MED_MAX_E"]
    data["sn_rms_e"] = tm["PA_TDS_SC_STAT_SN_RMS_E"]
    data["sn_threshold"] = tm["PA_TDS_SC_STAT_SN_THRESHOLD"]
    data["du_nr_impact"] = tm["PA_TDS_SC_STAT_DU_NR_IMPACT"]
    data["du_med_imp"] = tm["PA_TDS_SC_STAT_DU_MED_IMP"]
    data["wa_amp_max"] = tm["PA_TDS_SC_STAT_WA_AMP_MAX"]
    data["wa_amp_med"] = tm["PA_TDS_SC_STAT_WA_AMP_MED"]
    data["wa_rms"] = tm["PA_TDS_SC_STAT_WA_RMS"]
    data["wa_nr_events"] = tm["PA_TDS_SC_STAT_WA_NR_EVENTS"]
    data["wa_med_freq"] = tm["PA_TDS_SC_STAT_WA_MED_FREQ"]

    return data



import numpy as np

from poppy.core.tools.exceptions import print_exception

from roc.rpl.time import Time
from roc.rpl.packet_structure.data cimport Data
from roc.rpl.packet_structure.data import Data

from roc.idb.converters.cdf_fill import fill_records as set_fillval

from roc.rap.tasks.thr.science_tools import voltages
from roc.rap.tasks.thr.science_tools import input_setup
from roc.rap.tasks.thr.science_tools import do_analysis
from roc.rap.tasks.thr.science_tools import voltages_dtype
from roc.rap.tasks.thr.science_tools import input_setup_dtype
from roc.rap.tasks.thr.science_tools import do_analysis_dtype
from roc.rap.tasks.thr.hfr import extract_band
from roc.rap.tasks.thr.hfr import hfr_setup
from roc.rap.tasks.thr.hfr import hfr_dtype
from roc.rap.tasks.thr.hfr import hfr_sweep_setup
from roc.rap.tasks.thr.hfr import compute_frequency


class HFRCalibrationDecomError(Exception):
    """
    Error management for HFR calibration.
    """


cpdef decommute(group, task):
    cdef:
        Py_ssize_t i

    # get the size for the number of packets
    number = group["PA_THR_HFR_DATA_CNT"].shape[0]

    # create the array containing the data
    array = np.empty(number, dtype=create_dtype())
    array.fill(np.nan)

    # loop over packets
    for i in range(number):
        try:
            # size for number of unsigned int
            size = group["PA_THR_HFR_DATA_CNT"][i]

            # transform the data into byte array
            # store because of the garbage collector removing reference while
            # processing in cython
            byte = group[
                "PA_THR_HFR_DATA"
            ][i][:size].byteswap().newbyteorder().tobytes()
            data = Data(byte, len(byte))

            # now extract the data
            result, offset = extract_data(
                data,
                0,
                group["PA_THR_ACQUISITION_TIME"][i],
            )

            # check offset size
            if offset != size * 2:
                raise HFRCalibrationDecomError(
                    "Expected a size of {0} bytes but have {1}".format(
                        size * 2,
                        offset,
                    )
                )

            # store the data
            array[i] = result
        except:
            message = print_exception()
            task.exception(message)

    # get the number of records
    number = band_record_number(array)

    # create an array of tnr records
    records = np.empty(number, dtype=hfr_dtype())
    set_fillval(records)

    # fill the array with the records
    records = fill_records(array, records)

    return records


cdef fill_records(array, records):
    cdef:
        Py_ssize_t i
        unsigned long nblocks
        unsigned long counter

    # get number of blocks
    nblocks = array.shape[0]

    # init the counter for placing data
    counter = 0
    # loop over fields
    for i in range(nblocks):
        if array["hfr_setup"]["hf1"][i] == 1:
            counter = set_record(
                records,
                array,
                counter,
                i,
                "hf1",
                0,
                "hf1_step",
                "hf1_size",
            )
        if array["hfr_setup"]["hf2"][i] == 1:
            counter = set_record(
                records,
                array,
                counter,
                i,
                "hf2",
                1,
                "hf2_step",
                "hf2_size",
            )

    return records


cdef set_record(records, array, counter, i, band, index, step, size):
    cdef:
        double epoch
        unsigned long frequency
        unsigned char level
        double delta_time = 0
        Py_ssize_t j

    # TT2000 Epoch times
    epoch = <double>Time().obt_to_utc(
        np.array([array["acquisition_time"][i][:2]]), to_tt2000=True)

    for level in range(9):
        for j in range(array["hfr_sweep"][step][i]):
            records["epoch"][counter] = epoch
            records["acquisition_time"][counter] = array["acquisition_time"][i][:2]
            records["scet"][counter] = Time.cuc_to_scet(
            np.array([array["acquisition_time"][i][:2]]))
            records["synchro_flag"][counter] = array[
                "acquisition_time"
            ][i][2]
            frequency = compute_frequency(
                array["hfr_setup"]["initial_frequency"][i],
                j,
                array["hfr_sweep"]["hf1_size"][i],
                array["hfr_sweep"]["hf1_step"][i],
                array["hfr_sweep"]["hf2_size"][i],
                band,
            )
            records["frequency"][counter] = frequency
            records["calibration_level"][counter] = level
            records["survey_mode"][counter] = 2
            records["average_nr"][counter] = array["hfr_setup"]["av"][i]
            records["channel_status"][counter] = (
                array["do_analysis"]["ch1"][i],
                array["do_analysis"]["ch2"][i],
            )
            records["front_end"][counter] = array["input_setup"]["fe"][i]
            records["sensor_config"][counter] = (
                array["input_setup"]["sensor_tnr_1"][i],
                array["input_setup"]["sensor_tnr_2"][i],
             )

            records["temperature"][counter] = (
                array["temperature_pcb"][i],
                array["voltages"]["minus5"][i],
                array["voltages"]["plus5"][i],
                array["voltages"]["plus12"][i],
            )
            records["hfr_band"][counter] = index
            records["agc1"][counter] = array["data"][
                str(level)
            ]["ch1"][band][i][j]
            records["agc2"][counter] = array["data"][
                str(level)
            ]["ch2"][band][i][j]
            counter += 1

    return counter


cpdef unsigned long band_record_number(array):
    """
    Get the number of records for each band.
    """
    cdef:
        tuple nnormal
        unsigned long number = 0
    filter1 = array["hfr_setup"]["hf1"] == 1
    filter2 = array["hfr_setup"]["hf2"] == 1

    nnormal = (
        int(array["hfr_sweep"]["hf1_step"][filter1[:]].sum() * 9),
        int(array["hfr_sweep"]["hf2_step"][filter2[:]].sum() * 9),
    )
    number += sum(nnormal)

    return number



cdef tuple extract_data(Data data, Py_ssize_t offset, time):
    """
    Extract the data and insert it into the data structure of the packet.
    """
    cdef:
        unsigned short cmd_ctr
        unsigned short n1, n2
        unsigned int bit_count
        tuple voltages_cmd
        tuple hfr_setup_cmd
        tuple input_setup_cmd
        tuple hfr_sweep_cmd
        tuple do_analysis_cmd
        tuple res
        list ch1, ch2
        Py_ssize_t i

    # read command control
    cmd_ctr = data.u8(offset, 0, 8)
    offset += 1

    # read voltages
    voltages_cmd = voltages(data, offset)
    offset += 3

    # read tnr setup
    hfr_setup_cmd = hfr_setup(data, offset)
    offset += 2

    # read input setup
    input_setup_cmd = input_setup(data, offset)
    offset += 2

    # read hfr sweep setup
    hfr_sweep_cmd = hfr_sweep_setup(data, offset)
    n1 = hfr_sweep_cmd[2]
    n2 = hfr_sweep_cmd[0]
    offset += 4

    # read do analysis command
    offset += 1
    do_analysis_cmd = do_analysis(data, offset)
    offset += 1

    # initialize bit count
    bit_count = 0

    # init containers for values of the calibration
    ch1, ch2 = [], []

    # loop over calibration levels
    for i in range(9):
        # extract bands
        res, bit_count = extract_band(
            do_analysis_cmd[4],
            hfr_setup_cmd[3],
            hfr_setup_cmd[2],
            n1,
            n2,
            data,
            offset,
            bit_count,
        )
        ch1.append(res)
        res, bit_count = extract_band(
            do_analysis_cmd[3],
            hfr_setup_cmd[3],
            hfr_setup_cmd[2],
            n1,
            n2,
            data,
            offset,
            bit_count,
        )
        ch2.append(res)

    # compute the new offset for the block
    offset += bit_count // 8

    # align
    if offset % 2 != 0:
        offset += 1

    # tuple of values to add them to the numpy array
    return (
        cmd_ctr, voltages_cmd, hfr_setup_cmd,
        input_setup_cmd, hfr_sweep_cmd, do_analysis_cmd,
        tuple([(ch1[i], ch2[i]) for i in range(9)]), time
    ), offset


cpdef list create_dtype():
    """
    Create dtype of numpy for the special packet.
    """
    return [
        ("temperature_pcb", "uint8"),
        ("voltages", voltages_dtype()),
        ("hfr_setup", setup_dtype()),
        ("input_setup", input_setup_dtype()),
        ("hfr_sweep", hfr_sweep_setup_dtype()),
        ("do_analysis", do_analysis_dtype()),
        ("data", block_data_dtype()),
        ("acquisition_time", "uint32", 3),
    ]


cpdef list channel_dtype():
    return [
        ("hf1", "uint16", 512), ("hf2", "uint16", 512)
    ]


cpdef list block_data_dtype():
    return [
        (
            "{0}".format(i),
            [
                ("ch1", channel_dtype()),
                ("ch2", channel_dtype())
            ],
        ) for i in range(9)
    ]


cpdef list hfr_sweep_setup_dtype():
    return [
        ("hf2_step", "uint16"), ("hf2_size", "uint8"),
        ("hf1_step", "uint16"), ("hf1_size", "uint8"),
    ]


cpdef list setup_dtype():
    return [
        ("sw", "uint8"), ("initial_frequency", "uint16"),
        ("hf2", "uint8"), ("hf1", "uint8"), ("av", "uint8")
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Module to deal with the RPW L1 BIAS CDF file production."""

import os
import csv

from roc.rap.tasks.bia.current import set_current
from roc.rap.tasks.bia.sweep import set_sweep

__all__ = ["extract_bia_current", "extract_bia_sweep"]

def extract_bia_current(l0, task):
    """
    Extract BIAS current data from L0 file.

    :param l0:
    :param task:
    :return:
    """

    data = set_current(l0, task)

    return data

def extract_bia_sweep(l0, task):
    """
    Extract Bias sweep data from L0 file.

    :param l0: h5py object containing l0 file data
    :param task: pipeline task instance
    :return: Bias sweep data as a numpy ndarray
    """

    # Get LFR data during Bias sweep
    data = None
    if 'TM_DPU_SCIENCE_BIAS_CALIB' in l0['TM']:
        data = set_sweep(l0['TM']['TM_DPU_SCIENCE_BIAS_CALIB']['source_data'], task, 'F3')
    elif 'TM_DPU_SCIENCE_BIAS_CALIB_LONG' in l0['TM']:
        data = set_sweep(l0['TM']['TM_DPU_SCIENCE_BIAS_CALIB_LONG']['source_data'], task, 'LONG_F3')

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import timedelta
from copy import deepcopy

import numpy as np
import pandas as pd
from sqlalchemy import asc, and_
from sqlalchemy.orm.exc import NoResultFound

from poppy.core.db.connector import Connector
from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.dingo.models.data import HfrTimeLog
from roc.dingo.tools import actual_sql

from roc.rap.constants import PIPELINE_DATABASE, UINT32_MAX_ENCODING

__all__ = ['get_hfr_time_log', 'get_hfr_delta_times',
           'merge_coarse_fine']


@Connector.if_connected(PIPELINE_DATABASE)
def get_hfr_time_log(mode, start_time, end_time,
                     model=HfrTimeLog):
    """
    Query pipeline.hfr_time_log table in database

    :param mode: Mode of HFR receiver (0=NORMAL, 1=BURST)
    :param start_time: minimal acquisition time to query (datetime object)
    :param end_time: maximal acquisition time to query (datetime object)
    :param model: database table class
    :return:
    """
    # get a database session
    session = Connector.manager[PIPELINE_DATABASE].session
    # Set filters
    logger.debug(f'Connecting {PIPELINE_DATABASE} database to retrieve entries from pipeline.hfr_time_log table ...')
    filters = []
    filters.append(model.mode == mode)  # Get data for current mode
    filters.append(model.acq_time >= str(start_time))
    filters.append(model.acq_time <= str(end_time))
    # Join query conditions with AND
    filters = and_(*filters)

    # Query database (returns results as a pandas.DataFrame object)
    try:
        query = session.query(model).filter(
            filters).order_by(asc(HfrTimeLog.acq_time))
        results = pd.read_sql(query.statement, session.bind)
    except NoResultFound:
        logger.exception(f'No result for {actual_sql(query)}')
        results = pd.DataFrame()

    return results


def get_hfr_delta_times(mode, acq_time_min,
                        duration=48):
    """
    Retrieve delta_times values in the pipeline.hfr_time_log table

    :param mode: HFR receiver mode (0=NORMAL, 1=BURST)
    :param acq_time_min: Minimal acquisition time to query (tuple with coarse and fine parts)
    :param duration: Time range to query will be [acq_start_time, acq_start_time + duration] with duration in hours. default is 24.
    :return: pandas.DataFrame with database query results
    """
    delta_times = [[], []]
    # Convert input acquisition time (coarse, fine) into
    # datetime object
    start_time = Time.cuc_to_datetime(acq_time_min)[0] - timedelta(minutes=1)
    end_time = start_time + timedelta(hours=duration)

    # Query database
    try:
        results = get_hfr_time_log(mode, start_time, end_time)
    except Exception as e:
        logger.exception('Cannot query pipeline.hfr_time_log table')
        raise e

    if results.shape[0] > 0:
        # If no empty dataframe, then make sure to have valid
        # delta times values
        results['delta_time1'] = results['delta_time1'].apply(valid_delta_time)
        results['delta_time2'] = results['delta_time2'].apply(valid_delta_time)

        # Add a variable containing merged coarse/fine parts of
        # acquisition time
        results['coarse_fine'] = results.apply(
            lambda row: merge_coarse_fine(
                row['coarse_time'], row['fine_time']),
            axis=1)

    return results


def valid_delta_time(delta_times):
    """
    Check if delta times values are valid for the input hfr_time_log row
    (i.e., first delta_time1 values for each record should be 0
        # and delta_times values should increase monotonically)

    :param delta_times1: hfr_time_log.delta_time 1 or 2 values to check
    :return: row with updates (if needed)
    """
    # Initialize output
    new_delta_times = deepcopy(delta_times)

    # Initialize byte offset
    offset = np.uint64(0)

    # Initialize previous value to compare
    prev_val = 0

    # Loop over packets
    for key, val in delta_times.items():
        # Loop over delta time values for current packet
        for j in range(0, len(val)):
            # Define index of value to compare
            if j > 0:
                prev_val = val[j - 1]

            # If previous delta time value is upper or equal than current value
            # then the increase is not monotone,
            # which indicates the max encoding value is reached
            if val[j] + offset < prev_val + offset:
                # In this case, increase offset
                offset += np.uint64(UINT32_MAX_ENCODING)
                logger.warning(f'Maximal encoding value reached for 32-bits delta time (offset={offset})')
            # Save delta time value with correct offset
            new_delta_times[key][j] = np.uint64(val[j]) + offset
            prev_val = val[j]

    return new_delta_times


def merge_coarse_fine(coarse, fine):
    """
    return input coarse and fine parts of CCSDS CUC time
    as an unique integer

    :param coarse: coarse part of CUC time
    :param fine: fine part of CUC time
    :return: resulting unique integer
    """
    return int(coarse * 100000 + fine)


def found_delta_time(group, delta_times, packet_times, i, message):
    """
    Get HF1 / HF2 delta times values for current HFR science packet

    :param group: h5.group containing HFR science packet source_data
    :param delta_times: pandas.DataFrame storing hfr_time_log table data
    :param packet_times: list of packet (creation) times (coarse, fine, sync)
    :param i: index of the current HFR science packet in the L0 file
    :param message: string containing HFR receiver mode ('normal' or 'burst')
    :return: delta_time1 and delta_time2 values found in delta_times for current packet
    """
    # Get row index where condition is fulfilled
    where_packet = (merge_coarse_fine(group['PA_THR_ACQUISITION_TIME'][i][0], group['PA_THR_ACQUISITION_TIME'][i][1]) ==
                    delta_times['coarse_fine'])
    # Must be only one row
    if delta_times[where_packet].shape[0] > 1:
        raise ValueError(f'Wrong number of hfr_time_log entries found for HFR {message} packet at {packet_times[i][:2]}: '
                         f'1 expected but {delta_times[where_packet].shape[0]} found! '
                         f'(acq. time is {group["PA_THR_ACQUISITION_TIME"][i][:2]})')
    # Must be at least one row
    elif delta_times[where_packet].shape[0] == 0:
        raise ValueError(f'No hfr_time_log entry found for HFR {message} packet #{i} at {packet_times[i][:2]} '
                         f'(acq. time is {group["PA_THR_ACQUISITION_TIME"][i][:2]})')
    else:
        # Else extract delta_time1 and delta_time2 values found
        delta_time1 = delta_times['delta_time1'][where_packet].reset_index(drop=True)[0][
            str(merge_coarse_fine(packet_times[i][0], packet_times[i][1]))]
        delta_time2 = delta_times['delta_time2'][where_packet].reset_index(drop=True)[0][
            str(merge_coarse_fine(packet_times[i][0], packet_times[i][1]))]

    return delta_time1, delta_time2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os.path as osp

from poppy.core.generic.paths import Paths

__all__ = ["paths"]

# root directory of the module
_ROOT_DIRECTORY = osp.abspath(
    osp.join(
        osp.dirname(__file__),
        osp.pardir,
    )
)

# create a path object that can be used to get some common path in the module
paths = Paths(_ROOT_DIRECTORY)






import numpy as np
cimport numpy as np

from roc.rpl.packet_structure.data import Data
from roc.rpl.packet_structure.data cimport Data


DURATION = {
    (0, 0): 0.35,
    (0, 1): 0.21,
    (0, 2): 0.17,
    (0, 3): 0.16,
    (1, 0): 0.63,
    (1, 1): 0.39,
    (1, 2): 0.33,
    (1, 3): 0.31,
    (2, 0): 1.16,
    (2, 1): 0.74,
    (2, 2): 0.63,
    (2, 3): 0.6,
    (3, 0): 2.14,
    (3, 1): 1.35,
    (3, 2): 1.25,
    (3, 3): 1.15,
}


cpdef list tnr_dtype():
    """
    Numpy array dtype for the TNR L1 survey data.
    """
    # Name convention is lower case
    return [
        ("epoch", "float64"),
        ("scet", "float64"),
        ("acquisition_time", "uint32", 2),
        ("synchro_flag", "uint8"),
        ("quality_flag", "uint8"),
        ("quality_bitmask", "uint8"),
        ("sweep_num", "uint32"),
        ("measurement_duration", "float64"),
        ("ticks_nr", "int64"),
        ("delta_time", "float64"),
        ("survey_mode", "uint8"),
        ("calibration_level", "uint8"),
        ("average_nr", "uint8"),
        ("auto_cross_status", "uint8", 2),
        ("channel_status", "uint8", 2),
        ("front_end", "uint8"),
        ("sensor_config", "uint8", 2),
        ("rpw_status", "uint8", 15),
        ("temperature", "uint8", 4),
        ("tnr_band", "uint8"),
        ("agc1", "uint16"),
        ("agc2", "uint16"),
        ("frequency", "uint32", 32),
        ("auto1", "uint16", 32),
        ("auto2", "uint16", 32),
        ("cross_r", "uint16", 32),
        ("cross_i", "uint16", 32),
    ]


cpdef tuple tnr_setup(Data data, unsigned int start):
    """
    Return tuple of values from the TNR setup command.
    """
    cdef:
        unsigned char A
        unsigned char B
        unsigned char C
        unsigned char D
        unsigned char au
        unsigned char cr
        unsigned char av

    # read parameters
    a = data.u8(start, 7, 1)
    b = data.u8(start, 6, 1)
    c = data.u8(start, 5, 1)
    d = data.u8(start, 4, 1)
    au = data.u8(start, 3, 1)
    cr = data.u8(start, 2, 1)
    av = data.u8(start, 0, 2)

    return av, cr, au, d, c, b, a


cpdef tuple delta_times(Data data, unsigned int start):
    """
    Read delta times command.
    """
    cdef:
        long long a
        long long b
        long long c
        long long d

    a = <long long>data.u32(start, 0, 32)
    b = <long long>data.u32(start + 4, 0, 32)
    c = <long long>data.u32(start + 8, 0, 32)
    d = <long long>data.u32(start + 12, 0, 32)

    return a, b, c, d


cpdef tuple extract_band(
    unsigned char band,
    unsigned char ch1,
    unsigned char ch2,
    unsigned char auto,
    unsigned char cross,
    Data data,
    unsigned int offset,
    unsigned int bit_count,
):
    """
    Extract the band, returning fake data if not selected.
    """
    cdef:
        Py_ssize_t i
        unsigned short ch1_agc_value
        unsigned short ch2_agc_value
        np.ndarray[np.uint16_t, ndim=1] ch1_auto_value = np.zeros(32, dtype="uint16")
        np.ndarray[np.uint16_t, ndim=1] ch2_auto_value = np.zeros(32, dtype="uint16")
        np.ndarray[np.uint16_t, ndim=2] cross_value = np.zeros((32, 2), dtype="uint16")

    # By default, fill values with FILLVAL (see CDF ISTP standards)
    ch1_auto_value[:] = 65535
    ch2_auto_value[:] = 65535
    cross_value[:,:] = 65535

    # if band is activated
    if band == 1:
        # read the channel #1 first
        if ch1 == 0:
            ch1_agc_value = data.u16(
                offset + bit_count // 8,
                bit_count % 8,
                12,
            )
            bit_count += 12
            if auto == 1:
                for i in range(32):
                    ch1_auto_value[i] = data.u16(
                        offset + bit_count // 8,
                        bit_count % 8,
                        12,
                    )
                    bit_count += 12
        else:
            ch1_agc_value = 65535

        # Then read channel #2
        if ch2 == 0:
            ch2_agc_value = data.u16(
                offset + bit_count // 8,
                bit_count % 8,
                12,
            )
            bit_count += 12
            if auto == 1:
                for i in range(32):
                    ch2_auto_value[i] = data.u16(
                        offset + bit_count // 8,
                        bit_count % 8,
                        12,
                    )
                    bit_count += 12
        else:
            ch2_agc_value = 65535

        # if cross and all channels
        if cross == 1 and ch1 == 0 and ch2 == 0:
            for i in range(32):
                cross_value[i, 0] = data.u16(
                    offset + bit_count // 8,
                    bit_count % 8,
                    12,
                )
                cross_value[i, 1] = data.u16(
                    offset + (bit_count + 384) // 8,
                    (bit_count + 384) % 8,
                    12,
                )
                bit_count += 12
            bit_count += 384

    return (
        (
            (ch1_agc_value, ch1_auto_value),
            (ch2_agc_value, ch2_auto_value),
            cross_value,
        ), bit_count
    )


cpdef tuple extract_calibration_band(
    unsigned char band,
    unsigned char ch1,
    unsigned char ch2,
    unsigned char auto,
    unsigned char cross,
    Data data,
    unsigned int offset,
    unsigned int bit_count,
):
    """
    Extract the band, returning fake data if not selected.
    """
    cdef:
        Py_ssize_t i
        unsigned short ch1_value
        unsigned short ch2_value
        np.ndarray[np.uint16_t, ndim=1] ch1_auto_value = np.zeros(8, dtype="uint16")
        np.ndarray[np.uint16_t, ndim=1] ch2_auto_value = np.zeros(8, dtype="uint16")
        np.ndarray[np.uint16_t, ndim=2] cross_value = np.zeros((8, 2), dtype="uint16")

    # if band is activated
    if band == 1:
        # read the first channel
        if ch1 == 0:
            ch1_value = data.u16(
                offset + bit_count // 8,
                bit_count % 8,
                12,
            )
            bit_count += 12
            if auto == 1:
                for i in range(8):
                    ch1_auto_value[i] = data.u16(
                        offset + bit_count // 8,
                        bit_count % 8,
                        12,
                    )
                    bit_count += 12
        else:
            ch1_value = 0

        # second channel
        if ch2 == 0:
            ch2_value = data.u16(
                offset + bit_count // 8,
                bit_count % 8,
                12,
            )
            bit_count += 12
            if auto == 1:
                for i in range(8):
                    ch2_auto_value[i] = data.u16(
                        offset + bit_count // 8,
                        bit_count % 8,
                        12,
                    )
                    bit_count += 12
        else:
            ch2_value = 0

        # if cross and all channels
        if cross == 1 and ch1 == 0 and ch2 == 0:
            for i in range(8):
                cross_value[i, 0] = data.u16(
                    offset + bit_count // 8,
                    bit_count % 8,
                    12,
                )
                cross_value[i, 1] = data.u16(
                    offset + (bit_count + 96) // 8,
                    (bit_count + 96) % 8,
                    12,
                )
                bit_count += 12
            bit_count += 96

    return (
        (
            (ch1_value, ch1_auto_value),
            (ch2_value, ch2_auto_value),
            cross_value,
        ), bit_count
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records

from roc.rap.tasks.utils import sort_data

__all__ = ['HIST1D', 'set_hist1d']

# Numpy array dtype for TDS histo1D L1 survey data
# Name convention is lower case

HIST1D = [
    ('epoch', 'uint64'),
    ('scet', 'float64'),
    ('acquisition_time', ('uint32', 2)),
    ('synchro_flag', 'uint8'),
    ('quality_flag', 'uint8'),
    ('quality_bitmask', 'uint8'),
    ('survey_mode', 'uint8'),
    ('bia_status_info', ('uint8', 6)),
    ('sampling_rate', 'float32'),
    ('rpw_status_info', ('uint8', 8)),
    ('channel_status_info', ('uint8', 4)),
    ('input_config', 'uint32'),
    ('snapshot_len', 'uint8'),
    ('hist1d_id', 'uint8'),
    ('hist1d_param', 'uint8'),
    ('hist1d_axis', 'uint8'),
    ('hist1d_col_time', 'uint16'),
    ('hist1d_out', 'uint16'),
    ('hist1d_bins', 'uint16'),
    ('hist1d_counts', ('uint16', 256)),
]

# TDS HIST1D sampling rate in Hz
SAMPLING_RATE_HZ = [65534.375, 131068.75, 262137.5, 524275., 2097100.]

def set_hist1d(l0, task):

    # name of the packet
    name = 'TM_TDS_SCIENCE_NORMAL_HIST1D'
    logger.debug('Get data for ' + name)

    # get the HIST1D packet data
    if name in l0['TM']:
        tm = l0['TM'][name]['source_data']
    else:
        # rest of the code inside the with statement will not be called
        return np.empty(
            0,
            dtype=HIST1D,
        )

    # Sort input packets by ascending acquisition times
    tm, __ = sort_data(tm, (
        tm['PA_TDS_ACQUISITION_TIME'][:,1],
        tm['PA_TDS_ACQUISITION_TIME'][:,0],
                    ))

    # get the number of packets
    size = tm['PA_TDS_ACQUISITION_TIME'].shape[0]

    # create the data array with good size
    data = np.empty(size, dtype=HIST1D)
    fill_records(data)

    # set the data for stat, directly a copy of the content of parameters
    data['epoch'][:] = Time().obt_to_utc(tm['PA_TDS_ACQUISITION_TIME'][:, :2], to_tt2000=True)
    data['acquisition_time'][:, :] = tm['PA_TDS_ACQUISITION_TIME'][:, :2]
    data['synchro_flag'][:] = tm['PA_TDS_ACQUISITION_TIME'][:, 2]
    data['scet'][:] = Time.cuc_to_scet(data['acquisition_time'])
    data['survey_mode'][:] = 3

    # bias info
    data['bia_status_info'][:, :] = np.vstack(
        (
            tm['PA_BIA_ON_OFF'],
            tm['PA_BIA_MODE_BIAS3_ENABLED'],
            tm['PA_BIA_MODE_BIAS2_ENABLED'],
            tm['PA_BIA_MODE_BIAS1_ENABLED'],
            tm['PA_BIA_MODE_HV_ENABLED'],
            tm['PA_BIA_MODE_MUX_SET'],
        )
    ).T

    data['channel_status_info'][:, :] = np.vstack(
        (
        tm['PA_TDS_STAT_ADC_CH1'],
        tm['PA_TDS_STAT_ADC_CH2'],
        tm['PA_TDS_STAT_ADC_CH3'],
        tm['PA_TDS_STAT_ADC_CH4'],
        )
    ).T

    # rpw status
    data['rpw_status_info'][:, :] = np.vstack(
        (
            tm['PA_TDS_THR_OFF'],
            tm['PA_TDS_LFR_OFF'],
            tm['PA_TDS_ANT1_OFF'],
            tm['PA_TDS_ANT2_OFF'],
            tm['PA_TDS_ANT3_OFF'],
            tm['PA_TDS_SCM_OFF'],
            tm['PA_TDS_HF_ART_MAG_HEATER'],
            tm['PA_TDS_HF_ART_SCM_CALIB'],
        )
    ).T

    # input config
    data['input_config'] = tm['PA_TDS_SWF_HF_CH_CONF']

    # hist parameters
    data['snapshot_len'] = tm['PA_TDS_SNAPSHOT_LEN']
    data['hist1d_id'] = tm['PA_TDS_SC_HIST1D_ID']
    data['hist1d_param'] = tm['PA_TDS_SC_HIST1D_PARAM']
    data['hist1d_axis'] = tm['PA_TDS_SC_HIST1D_BINS']
    data['hist1d_col_time'] = tm['PA_TDS_SC_HIST1D_COLL_TIME']
    data['hist1d_out'] = tm['PA_TDS_SC_HIST1D_OUT']
    data['hist1d_bins'] = tm['PA_TDS_SC_HIST1D_NUM_BINS']

    for i in range(size):
        # sampling rate
        data['sampling_rate'][i] = SAMPLING_RATE_HZ[tm['PA_TDS_STAT_SAMP_RATE'][i]]
        # HIST1D count
        ncount = len(tm['PA_TDS_SC_HIST1D_COUNTS'][i, :])
        data['hist1d_counts'][i, 0:ncount] = tm['PA_TDS_SC_HIST1D_COUNTS'][i, :]

    return data



import numpy as np
cimport numpy as np
from libcpp cimport bool as bool_t

from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records as filler
from poppy.core.tools.exceptions import print_exception
from roc.rpl.packet_structure.data cimport Data
from roc.rpl.packet_structure.data import Data
from libc.stdint cimport uint64_t

from roc.rap.tasks.thr.hfr_time_log import found_delta_time
from roc.rap.tasks.thr.science_tools import rpw_status
from roc.rap.tasks.thr.science_tools import temperatures_command
from roc.rap.tasks.thr.science_tools import input_setup
from roc.rap.tasks.thr.science_tools import do_analysis
from roc.rap.tasks.thr.science_tools import rpw_status_dtype
from roc.rap.tasks.thr.science_tools import temperatures_dtype
from roc.rap.tasks.thr.science_tools import input_setup_dtype
from roc.rap.tasks.thr.science_tools import do_analysis_dtype
from roc.rap.tasks.thr.science_tools import average_nr
from roc.rap.tasks.thr.hfr import extract_band
from roc.rap.tasks.thr.hfr import hfr_setup
from roc.rap.tasks.thr.hfr import hfr_sweep_setup
from roc.rap.tasks.thr.hfr import delta_times
from roc.rap.tasks.thr.hfr import hfr_dtype
from roc.rap.tasks.thr.hfr import compute_frequency
from roc.rap.tasks.thr.hfr import compute_tempo
from roc.rap.tasks.thr.hfr import hfr_band_index
from roc.rap.tasks.thr.hfr_list import HfrListMode
from roc.rap.constants import THR_TICK_NSEC

cpdef decommute_burst(group, task, packet_times, delta_times, mode):
    return decommute(group, task, packet_times, delta_times, mode)


cpdef decommute_normal(group, task, packet_times, delta_times, mode):
    return decommute(group, task, packet_times, delta_times, mode)


cdef decommute(group, task, packet_times, delta_times, mode):
    cdef:
        Py_ssize_t i
        Py_ssize_t j
        long number
        long packet_number

    # Initialize failed flag
    has_failed = False

    # select the mode
    if mode == 0:
        message = "normal"
        cnt_param = "PA_THR_N_BLOCK_CNT"
    else:
        message = "burst"
        cnt_param = "PA_THR_B_BLOCK_CNT"

    # get the size for the number of packets
    number = np.sum(group[cnt_param])

    # create the array containing the data
    array = np.empty(number, dtype=create_dtype())
    array.fill(np.nan)

    # the number of packets
    packet_number = group[cnt_param].shape[0]

    # loop over packets
    counter = 0
    for i in range(packet_number):
        try:
            # size of the number of unsigned int
            size = group["PA_THR_HFR_DATA_CNT"][i]

            # number of blocks
            nblocks = group[cnt_param][i]

            # Get delta times values for current packet
            try:
                delta_time1, delta_time2 = found_delta_time(
                    group, delta_times, packet_times, i, message)
            except:
                raised_message = print_exception()
                task.exception(raised_message)
                has_failed = True
                break

            # transform the data into byte array
            # store because of the garbage collector removing reference while
            # processing in cython
            byte = group[
                "PA_THR_HFR_DATA"
            ][i][:size].byteswap().newbyteorder().tobytes()
            data = Data(byte, len(byte))

            # init offset
            offset = 0

            # loop over blocks
            for j in range(nblocks):
                # now extract the data
                array[counter], offset = extract_data(
                    data,
                    offset,
                    group["PA_THR_ACQUISITION_TIME"][i],
                    delta_time1[j],
                    delta_time2[j],
                )

                # increment counter
                counter += 1

            # check offset size
            if offset != size * 2:
                raise Exception(
                    "Bad decommutation for HFR " + message + " mode\n" +
                    "Expected {0} got {1} bytes".format(2 * size, offset)
                )
        except:
            # if problem with a packet, continue the process
            raised_message = print_exception()
            task.exception(raised_message)

    if not has_failed:
        # get the number of records
        number = band_record_number(array)

        # create an array of hfr records
        records = np.empty(number, dtype=hfr_dtype())
        filler(records)

        # fill the array with the records
        records = fill_records(array, records, mode)
    else:
        raise Exception(f'Processing HFR {message} has failed')

    return records


cdef fill_records(array, records, mode):
    cdef:
        Py_ssize_t i
        long nblocks
        unsigned long counter

    # get number of blocks
    nblocks = array.shape[0]

    # init the counter for placing data
    counter = 0

    # loop over blocks of data
    for i in range(nblocks):

        if array["hfr_setup"]["hf1"][i] == 1:

            # Fill output records with packet parameters values
            counter = set_record(
                records,
                array,
                mode,
                counter,
                i,
                "hf1",
                "hf1_step",
                "hf1_size",
            )
        if array["hfr_setup"]["hf2"][i] == 1:

            # Fill output records with packet parameters values
            counter = set_record(
                records,
                array,
                mode,
                counter,
                i,
                "hf2",
                "hf2_step",
                "hf2_size",
            )

    return records

cdef unsigned long set_record(records, array, mode, counter, i, band,
                              step, size) except 1:
    cdef:
        long long epoch
        unsigned long long tempo
        unsigned long long measure
        unsigned long frequency
        unsigned long long delta_time_nsec
        double sample_time
        double scet
        Py_ssize_t j

    # CAVEATS: Due to HFR acquisition time behaviour
    # the exact Epoch time values cannot be fully known at this stage,
    # since we assume here the acquisition time is updated every new TM packet,
    # (which is not the case, see https://gitlab.obspm.fr/ROC/RCS/THR_CALBAR/-/issues/76)
    # The actual Epoch time values are hence set using l1_postpro task in FILM
    # plugin.

    # Get delta time for the current band and block in nanoseconds
    delta_time_nsec = <unsigned long long > ((array["deltas"][band][i]) * THR_TICK_NSEC)

    # Compute scet time in seconds
    scet = (Time.cuc_to_scet(
        np.array([array["acquisition_time"][i][:2]])) +  # acquisition time
        float(delta_time_nsec / 1000000000.)  # delta time in seconds
    )

    # Compute epoch time for the first sample of the sweep / block
    epoch = <long long > (Time().obt_to_utc(
        np.array([array["acquisition_time"][i][:2]]), to_tt2000=True) +  # acquisition time
        delta_time_nsec  # delta time in nanoseconds
    )

#    print(i, band, mode, array["acquisition_time"][i][:2], array["deltas"][band][i], array["delta_times"][band][i], delta_time_nsec, epoch,
#          str(Time().obt_to_utc(np.array([array["acquisition_time"][i][:2]]))),
#          str(Time().tt2000_to_utc(epoch)))

    # Process HFR MODE list (if found)
    if array["hfr_setup"]["sw"][i] == 1:
        try:
            logger.debug("Getting HFR LIST mode frequency values")

            frequency_list = HfrListMode().get_freq(Time().obt_to_utc(
                np.array([array["acquisition_time"][i][:2]]),
                to_datetime=True)[0], band, mode)
            if not frequency_list:
                raise ValueError(
                    'No frequency list has been returned from pipeline.tc_log table')
            if len(frequency_list) != array["hfr_sweep"][step][i]:
                raise ValueError('Number of frequencies in database ({0}) and in telemetry data ({1}) are incompatible!'.format(
                    len(frequency_list), array["hfr_sweep"][step][i]))
        except Exception as e:
            logger.exception('Cannot retrieve frequencies for HFR LIST mode')
            raise e

    # Loop over list of frequencies
    tempo = 0
    for j in range(array["hfr_sweep"][step][i]):
        if array["hfr_setup"]["sw"][i] == 1:
            frequency = frequency_list[j]
        else:
            # Get frequency value
            frequency = compute_frequency(
                array["hfr_setup"]["initial_frequency"][i],
                j,
                array["hfr_sweep"]["hf1_size"][i],
                array["hfr_sweep"]["hf1_step"][i],
                array["hfr_sweep"]["hf2_size"][i],
                band,
            )

        # Save acquisition time of the first sample in the packet
        records["acquisition_time"][counter] = array["acquisition_time"][i][:2]

        # Get the duration of the measurement
        # for the given HFR frequency in milliseconds
        measure = compute_tempo(frequency)
        # Get sample time for the given frequency in milliseconds
        sample_time = 0.5 * measure + tempo

        # Compute Epoch time for the current sample in nanoseconds
        records["epoch"][counter] = epoch + (sample_time) * 1000000
        records["scet"][counter] = scet + float(sample_time)

        tempo += measure
        records["sweep_num"][counter] = i
        records["frequency"][counter] = frequency
        records["sample_time"][counter] = sample_time * \
            1000.  # in microseconds
        records["synchro_flag"][counter] = array[
            "acquisition_time"
        ][i][2]

        records["ticks_nr"][counter] = array["delta_times"][band][i]
        records["delta_time"][counter] = < double > (delta_time_nsec) / 1000.  # in microseconds

        records["survey_mode"][counter] = mode
        records["sweep_mode"][counter] = array["hfr_setup"]["sw"][i]
        records["average_nr"][counter] = average_nr(
            array["hfr_setup"]["av"][i])
        records["channel_status"][counter] = (
            array["do_analysis"]["ch1"][i],
            array["do_analysis"]["ch2"][i],
        )
        records["front_end"][counter] = array["input_setup"]["fe"][i]
        records["sensor_config"][counter] = (
            array["input_setup"]["sensor_tnr_1"][i],
            array["input_setup"]["sensor_tnr_2"][i],
        )
        records["rpw_status"][counter][:] = array["rpw_status"][i].tolist()
        records["temperature"][counter] = array["temperatures"][i].tolist()
        records["hfr_band"][counter] = hfr_band_index(band)
        records["agc1"][counter] = array["data"]["ch1"][band][i][j]
        records["agc2"][counter] = array["data"]["ch2"][band][i][j]
        counter += 1

    return counter


cdef unsigned long band_record_number(array):
    """
    Get the number of records for each band.
    """
    cdef:
        tuple nnormal
        unsigned long number = 0

    filter1 = array["hfr_setup"]["hf1"] == 1
    filter2 = array["hfr_setup"]["hf2"] == 1

    nnormal = (
        int(array["hfr_sweep"]["hf1_step"][filter1[:]].sum()),
        int(array["hfr_sweep"]["hf2_step"][filter2[:]].sum()),
    )
    number += sum(nnormal)

    return number


cdef tuple extract_data(Data data, Py_ssize_t offset, time,
                        delta1, delta2):
    cdef:
        unsigned short cmd_ctr
        unsigned short bit_count
        unsigned short n1, n2
        tuple rpw_status_cmd
        tuple temperatures
        tuple hfr_setup_cmd
        tuple input_setup_cmd
        tuple hfr_sweep_cmd
        tuple do_analysis_cmd
        tuple deltas
        tuple ch1, ch2

    # read command control
    cmd_ctr = data.u16(offset, 0, 16)
    offset += 2

    # read rpw status
    rpw_status_cmd = rpw_status(data, offset)
    offset += 2

    # read temperatures
    temperatures = temperatures_command(data, offset)
    offset += 4

    # read tnr setup
    hfr_setup_cmd = hfr_setup(data, offset)
    offset += 2

    # get the value for the sweep mode
    sw = hfr_setup_cmd[0]

    # read input setup
    input_setup_cmd = input_setup(data, offset)
    offset += 2

    # read hfr sweep setup
    hfr_sweep_cmd = hfr_sweep_setup(data, offset)
    n1 = hfr_sweep_cmd[2]
    n2 = hfr_sweep_cmd[0]
    offset += 4

    # read do analysis command
    offset += 1
    do_analysis_cmd = do_analysis(data, offset)
    offset += 1

    # read delta times
    # (NOT USED: see https://gitlab.obspm.fr/ROC/RCS/THR_CALBAR/-/issues/76 for details)
    deltas = delta_times(data, offset)
    offset += 8

    # initialize bit count
    bit_count = 0

    # extract bands
    ch1, bit_count = extract_band(
        do_analysis_cmd[4],
        hfr_setup_cmd[3],
        hfr_setup_cmd[2],
        n1,
        n2,
        data,
        offset,
        bit_count,
    )
    ch2, bit_count = extract_band(
        do_analysis_cmd[3],
        hfr_setup_cmd[3],
        hfr_setup_cmd[2],
        n1,
        n2,
        data,
        offset,
        bit_count,
    )

    # compute the new offset for the block
    offset += bit_count // 8

    # align
    if offset % 2 != 0:
        offset += 1

    # tuple of values to add them to the numpy array
    return (
        cmd_ctr, rpw_status_cmd, temperatures, hfr_setup_cmd,
        input_setup_cmd, hfr_sweep_cmd, do_analysis_cmd, deltas,
        (ch1, ch2), time, (delta1, delta2)
    ), offset


cpdef list create_dtype():
    """
    Create dtype of numpy for the special packet.
    """
    return [
        ("cmd_ctr", "uint8"),
        ("rpw_status", rpw_status_dtype()),
        ("temperatures", temperatures_dtype()),
        ("hfr_setup", setup_dtype()),
        ("input_setup", input_setup_dtype()),
        ("hfr_sweep", hfr_sweep_setup_dtype()),
        ("do_analysis", do_analysis_dtype()),
        ("delta_times", delta_times_dtype()),
        ("data", block_data_dtype()),
        ("acquisition_time", "uint32", 3),
        ("deltas", deltas_dtype())
    ]


cpdef list channel_dtype():
    return [
        ("hf1", "uint16", 512), ("hf2", "uint16", 512)
    ]


cpdef list block_data_dtype():
    return [
        ("ch1", channel_dtype()), ("ch2", channel_dtype()),
    ]


cpdef list delta_times_dtype():
    return [
        ("hf1", "int64"), ("hf2", "int64"),
    ]

cpdef list deltas_dtype():
    return [
        ("hf1", "uint64"), ("hf2", "uint64"),
    ]

cpdef list hfr_sweep_setup_dtype():
    return [
        ("hf2_step", "uint16"), ("hf2_size", "uint8"),
        ("hf1_step", "uint16"), ("hf1_size", "uint8"),
    ]


cpdef list setup_dtype():
    return [
        ("sw", "uint8"), ("initial_frequency", "uint16"),
        ("hf2", "uint8"), ("hf1", "uint8"), ("av", "uint8")
    ]



import numpy as np
cimport numpy as np

from poppy.core.logger import logger

from roc.rpl.time import Time
from poppy.core.tools.exceptions import print_exception
from roc.idb.converters.cdf_fill import fill_records, CDFToFill
from libc.stdint cimport uint8_t
from libc.stdint cimport uint16_t
from libc.stdint cimport uint32_t
from libc.stdint cimport uint64_t

from roc.rap.tasks.tds.utils import count_swf
from roc.rap.tasks.utils import sort_data

__all__ = ["RSWF", "set_rswf"]

NSAMP_MAX_PER_REC = 32768

# Numpy array dtype for TDS LFM RSWF L1 survey data
# Name convention is lower case
RSWF = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", ("uint32", 2)),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("survey_mode", "uint8"),
    ("bia_status_info", ("uint8", 6)),
    ("lf_data_artefacts", ("uint8", 16)),
    ("input_config", ("uint8", 8)),
    ("rpw_status_info", ("uint8", 8)),
    ("snapshot_seq_nr", "uint16"),
    ("samps_per_ch", "uint64"),
    ("channel_status_info", ("uint8", 8)),
    ("sampling_rate", "float32"),
    ("waveform_data", ("int16", (8, NSAMP_MAX_PER_REC))),
]

# Sampling rate (fixed 32768 sps)
SAMP_RATE_ENUM = 32768.0

# Possible number of samples per snapshot (2**N with 9<N<18)
SAMP_PER_SNAPSHOT = 2**np.arange(9,18,1).astype(np.uint64)

# Max. Numer of WF samples in a packet
NSAMP_MAX_PER_PKT = 2032

# Max number of channels
NCHANNEL = 6


cpdef set_rswf(l0, task):
    cdef:
        Py_ssize_t i, j, k, l
        uint8_t counter_ch
        uint8_t nchannel
        uint32_t snapshot_size
        uint32_t samp_per_ch
        uint32_t samp_per_ch_max
        uint32_t samp_per_ch_miss
        uint32_t samp_per_snpht
        uint32_t snapshot_nsamp
        np.ndarray data
        np.ndarray index
        np.ndarray[np.uint8_t, ndim=1] pkt_cnt
        np.ndarray[np.uint8_t, ndim=1] pkt_nr
        np.ndarray[np.uint32_t, ndim=1] snapshot_nr
        np.ndarray[np.uint32_t, ndim=1] snapshot_counter
        np.ndarray[np.uint32_t, ndim=1] snapshot_id
        np.ndarray[np.uint32_t, ndim=1] snpht
        np.ndarray[np.uint32_t, ndim=1] samp_diff


    # name of the packet
    name = "TM_TDS_SCIENCE_LFM_RSWF"
    logger.debug("Get data for " + name)

    # get the LFM RSWF packet data
    if name in l0["TM"]:
        tm = l0["TM"][name]["source_data"]
    else:
        # rest of the code inside the with statement will not be called
        return np.empty(
            0,
            dtype=RSWF,
        )

    # Sort input packets by ascending acquisition times
    tm, __ = sort_data(tm, (
        tm["PA_TDS_LFM_RSWF_SNAPSHOT_NR"][:],
        tm["PA_TDS_ACQUISITION_TIME"][:,1],
        tm["PA_TDS_ACQUISITION_TIME"][:,0],
                    ))

    # Number of expected packets per snapshot
    pkt_cnt = tm["PA_TDS_LFM_RSWF_PKT_CNT"][:]

    # sequence number of the packet in the snapshot (from 1 to PA_TDS_SWF_PKT_CNT)
    pkt_nr = tm["PA_TDS_LFM_RSWF_PKT_NR"][:]

    # Index of the snapshot
    snapshot_nr = tm["PA_TDS_LFM_RSWF_SNAPSHOT_NR"][:].astype(np.uint32)

    # Number of data samples per channel for missing packet
    samp_per_ch_miss = tm["PA_TDS_LFR_SAMPS_PER_CH"][0]

    # Get number of snapshots
    snapshot_counter, snapshot_size = count_swf(snapshot_nr)

    # Get unique index of snapshots
    snapshot_id = np.unique(snapshot_counter).astype(np.uint32)

    # # create the data array with good size
    data = np.empty(snapshot_size, dtype=RSWF)
    fill_records(data)

    # Loop over snapshots
    logger.debug("Number of snapshots: {0}".format(snapshot_size))
    with nogil:
        for i in range(snapshot_size):
            with gil:
                snpht = np.where(snapshot_counter == snapshot_id[i])[0].astype(np.uint32)

                # Sort snapshot packets by increasing PA_TDS_LFM_RSWF_PKT_NR values
                snpht = snpht[np.argsort(pkt_nr[snpht[:]])]

                # Fill acquisition time
                data["acquisition_time"][i, :] = tm["PA_TDS_ACQUISITION_TIME"][snpht[0], :2]

                data["epoch"][i] = Time().obt_to_utc(data["acquisition_time"][i, :].reshape([1, 2])
                                              ,to_tt2000=True)

                # the flag for time synchronization
                data["synchro_flag"][i] = tm["PA_TDS_ACQUISITION_TIME"][snpht[0], 2]

                # scet
                data["scet"][i] = Time.cuc_to_scet(data["acquisition_time"][i, :].reshape([1, 2]))

                # survey mode to normal
                data["survey_mode"][i] = 0

                # Sampling rate (fixed to 32768 sps)
                data["sampling_rate"][i] = SAMP_RATE_ENUM

                data["bia_status_info"][i][:] = (
                            tm["PA_BIA_ON_OFF"][snpht[0]],
                            tm["PA_BIA_MODE_BIAS3_ENABLED"][snpht[0]],
                            tm["PA_BIA_MODE_BIAS2_ENABLED"][snpht[0]],
                            tm["PA_BIA_MODE_BIAS1_ENABLED"][snpht[0]],
                            tm["PA_BIA_MODE_HV_ENABLED"][snpht[0]],
                            tm["PA_BIA_MODE_MUX_SET"][snpht[0]],
                        )

                data["channel_status_info"][i][:] = (
                    tm["PA_TDS_LFM_RSWF_CH1"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_CH2"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_CH3"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_CH4"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_CH5"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_CH6"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_CH7"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_CH8"][snpht[0]],
                )

                data["rpw_status_info"][i][:] = (
                    tm["PA_TDS_LF_ART_THR_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_LFR_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_ANT1_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_ANT2_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_ANT3_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_SCM_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_MAG_HEATER"][snpht[0]],
                    tm["PA_TDS_LF_ART_SCM_CALIB"][snpht[0]],
                )

                data["lf_data_artefacts"][i][:] = (
                    tm["PA_TDS_LF_ART_CH1_OR"][snpht[0]],
                    tm["PA_TDS_LF_ART_CH2_OR"][snpht[0]],
                    tm["PA_TDS_LF_ART_CH3_OR"][snpht[0]],
                    tm["PA_TDS_LF_ART_CH4_OR"][snpht[0]],
                    tm["PA_TDS_LF_ART_CH5_OR"][snpht[0]],
                    tm["PA_TDS_LF_ART_CH6_OR"][snpht[0]],
                    tm["PA_TDS_LF_ART_CH7_OR"][snpht[0]],
                    tm["PA_TDS_LF_ART_CH8_OR"][snpht[0]],
                    tm["PA_TDS_LF_ART_MAG_HEATER"][snpht[0]],
                    tm["PA_TDS_LF_ART_SCM_CALIB"][snpht[0]],
                    tm["PA_TDS_LF_ART_ANT3_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_ANT2_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_ANT1_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_SCM_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_THR_OFF"][snpht[0]],
                    tm["PA_TDS_LF_ART_LFR_OFF"][snpht[0]],
                )

                data["input_config"][i][:] = (
                    tm["PA_TDS_LFM_RSWF_INPUT_CONF1"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_INPUT_CONF2"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_INPUT_CONF3"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_INPUT_CONF4"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_INPUT_CONF5"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_INPUT_CONF6"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_SINGLE_MODE"][snpht[0]],
                    tm["PA_TDS_LFM_RSWF_RESERVED"][snpht[0]],
                )

                data["snapshot_seq_nr"][i] = snapshot_id[i]

                # Max. number of data samples per channel
                samp_per_ch_max = np.uint64(NSAMP_MAX_PER_PKT / np.sum(data["channel_status_info"][i][:]))

                # Total number of samples per snapshot
                samp_diff = np.abs((samp_per_ch_max * np.uint32(pkt_cnt[snpht[0]]) - SAMP_PER_SNAPSHOT)).astype(np.uint32)
                samp_per_snpht = SAMP_PER_SNAPSHOT[samp_diff.argsort()[0]]

                # Loop on each pachet of the snapshot
                snapshot_nsamp = 0
                j = 0
                k = 0
                while k < pkt_cnt[snpht[0]]:

                    if j >= len(snpht):
                        j = len(snpht) - 1

                    # Check that packet counter k value + 1
                    # is the same than the packet sequence
                    # number pkt_nr[snpht[j]] in the current snapshot
                    if k + 1 != pkt_nr[snpht[j]]:
                        # ... If not, then there is a missing packet.
                        logger.warning("Packet #{0} [{2}] is missing for snapshot #{1}!".format(k+1, i, pkt_cnt[snpht[0]]))

                        # If the missing packet is the last packet...
                        if k+1 == pkt_cnt[snpht[0]]:
                            # compute the number of sample
                            # in the last packet
                            samp_per_ch_miss = samp_per_snpht - np.uint32(samp_per_ch_max * (pkt_cnt[snpht[0]] - 1))

                        # If the missing packet is the first packet...
                        elif k+1 == 1:
                            # If there is only 2 packets for this snapshot
                            if pkt_cnt[snpht[0]] == 2:
                                # Compute the number of sample
                                # in the first packet
                                samp_per_ch_miss = samp_per_snpht - samp_per_ch_max
                            else:
                                # ... Get the number of samples from the next
                                # valid packet (which should be
                                # not the last one)
                                samp_per_ch_miss = samp_per_ch_max

                        # Increment total number of samples
                        # for the current snapshot
                        # assuming that the number of missing samples
                        # is the last samp_per_ch value
                        snapshot_nsamp += samp_per_ch_miss

                        # Increment packet counter k and continue
                        # in order to reach the next packet
                        # sequence number pkt_nr[snpht[j]] in the
                        # snapshot
                        k += 1
                        continue

                    # Fill TDS SWF sample data
                    samp_per_ch = np.uint32(tm["PA_TDS_LFR_SAMPS_PER_CH"][snpht[j]])

                    # Update the samp_per_ch_miss value
                    samp_per_ch_miss = samp_per_ch

                    index = np.arange(0, samp_per_ch, 1)

                    counter_ch = 0
                    for l in range(NCHANNEL):
                        if data["channel_status_info"][i][l] == 1:
                            data["waveform_data"][i, l, snapshot_nsamp : snapshot_nsamp + samp_per_ch] = tm[
                                     "PA_TDS_LFM_RSWF_DATA_BLK"
                                 ][snpht[j], index + (samp_per_ch * counter_ch)]
                            counter_ch += 1

                    # Increment total number of samples per channel
                    # for the current snapshot
                    snapshot_nsamp += samp_per_ch
                    k += 1
                    j += 1

                # Perform some verifications before saving the snapshot
                try:
                    # Check that total number of samples does not exceed the number of expected max. records in the output file
                    if snapshot_nsamp > NSAMP_MAX_PER_REC:
                        raise Exception("Total number of samples cannot exceed {1}, but {0} found in snapshot #{2}!".format(snapshot_nsamp, NSAMP_MAX_PER_REC, snapshot_id[i]))

                    # Check that the total number of samples is consistent
                    # with what it is expected
                    if samp_per_snpht != snapshot_nsamp:
                        raise Exception("Unexpected total number of samples in snapshot #{0}! ({1} instead of {2})".format(snapshot_id[i], snapshot_nsamp, samp_per_snpht))
                except:
                    message = print_exception()
                    task.exception(message)
                    data["waveform_data"][i, :, :] = CDFToFill()(data["waveform_data"].dtype.name)
                else:
                    data["samps_per_ch"][i] = snapshot_nsamp
    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

from poppy.core.logger import logger

from roc.rpl.time import Time

from roc.idb.converters.cdf_fill import fill_records

# Numpy array dtype for BIAS sweep data
# Name convention is lower case
from roc.rap.tasks.utils import sort_data

SWEEP_DTYPE = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", "uint32", 2),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("bias_mode_mux_set", "uint8"),
    ("bias_mode_hv_enabled", "uint8"),
    ("bias_mode_bias1_enabled", "uint8"),
    ("bias_mode_bias2_enabled", "uint8"),
    ("bias_mode_bias3_enabled", "uint8"),
    ("bias_mode_bypass_probe1", "uint8"),
    ("bias_mode_bypass_probe2", "uint8"),
    ("bias_mode_bypass_probe3", "uint8"),
    ("bias_on_off", "uint8"),
    ("ant_flag", "uint8"),
    ("bw", "uint8"),
    ("sp0", "uint8"),
    ("sp1", "uint8"),
    ("r0", "uint8"),
    ("r1", "uint8"),
    ("r2", "uint8"),
    ("sampling_rate", "float32"),
    ("v", "float32", 3),
    ("bias_sweep_current", "float32", 3),
]

# Factor to convert LFR Bias sweep data voltage into volts, i.e., V_TM ~ V_Volt * 8000 * 1/17
# Provided by Yuri
BIA_SWEEP_TM2V_FACTOR = 1.0 / (8000 * 1.0/17.0)

# F3 = 16Hz
CWF_SAMPLING_RATE = 16.0

def set_sweep(l0, task, freq):

    # Get/create time instance
    time_instance = Time()

    # Get the number of packets
    size = l0["PA_LFR_ACQUISITION_TIME"].shape[0]

    if freq == "F3":
        blk_nr = l0["PA_LFR_CWF3_BLK_NR"][:]
    elif freq == "LONG_F3":
        blk_nr = l0["PA_LFR_CWFL3_BLK_NR"][:]
        freq = freq[-2:]
    else:
        logger.warning(f'Unknown frequency label {freq}')
        return np.empty(size, dtype=SWEEP_DTYPE)

    # Sort input packets by ascending acquisition times
    l0, __ = sort_data(l0, (
        l0["PA_LFR_ACQUISITION_TIME"][:,1],
        l0["PA_LFR_ACQUISITION_TIME"][:,0],
                    ))

    # Total number of samples
    nsamps = np.sum(blk_nr)

    # Sampling rate
    samp_rate = CWF_SAMPLING_RATE

    # Cadence in nanosec
    delta_t = 1.0e9 / samp_rate

    # create a record array with the good dtype and good length
    data = np.empty(nsamps, dtype=SWEEP_DTYPE)
    fill_records(data)

    # Loop over packets
    i = 0
    for packet in range(size):

        # Get the number of samples for the current packet
        nsamp = blk_nr[packet]

        # get the acquisition time and time synch. flag
        data["acquisition_time"][i:i + nsamp, :] = \
            l0["PA_LFR_ACQUISITION_TIME"][packet, :2]
        data["synchro_flag"][i:i + nsamp] = \
            l0["PA_LFR_ACQUISITION_TIME"][packet, 2]

        # Compute corresponding Epoch time
        epoch0 = time_instance.obt_to_utc(
            l0["PA_LFR_ACQUISITION_TIME"][packet, :2].reshape([1, 2]),
        to_tt2000=True)

        data["epoch"][i:i + nsamp] = epoch0 +  np.uint64(delta_t * np.arange(0, nsamp, 1))

        data["scet"][i:i + nsamp] = Time.cuc_to_scet(data["acquisition_time"][i:i + nsamp, :])

        data["bias_mode_mux_set"][i:i + nsamp] = \
            l0["PA_BIA_MODE_MUX_SET"][packet]
        data["bias_mode_hv_enabled"][i:i + nsamp] = \
            l0["PA_BIA_MODE_HV_ENABLED"][packet]
        data["bias_mode_bias1_enabled"][i:i + nsamp] = \
            l0["PA_BIA_MODE_BIAS1_ENABLED"][packet]
        data["bias_mode_bias2_enabled"][i:i + nsamp] = \
            l0["PA_BIA_MODE_BIAS2_ENABLED"][packet]
        data["bias_mode_bias3_enabled"][i:i + nsamp] = \
            l0["PA_BIA_MODE_BIAS3_ENABLED"][packet]
        data["bias_on_off"][i:i + nsamp] = l0["PA_BIA_ON_OFF"][packet]
        data["bw"][i:i + nsamp] = l0["SY_LFR_BW"][packet]
        data["sp0"][i:i + nsamp] = l0["SY_LFR_SP0"][packet]
        data["sp1"][i:i + nsamp] = l0["SY_LFR_SP1"][packet]
        data["r0"][i:i + nsamp] = l0["SY_LFR_R0"][packet]
        data["r1"][i:i + nsamp] = l0["SY_LFR_R1"][packet]
        data["r2"][i:i + nsamp] = l0["SY_LFR_R2"][packet]

        # Fill sampling rate
        data["sampling_rate"][i:i + nsamp] = CWF_SAMPLING_RATE

        # Electrical potential on each antenna 1, 2 and 3 (just need to be reshaped)
        data["v"][i:i + nsamp, 0] = raw_to_volts(l0["PA_LFR_SC_V_" + freq][packet, :])
        data["v"][i:i + nsamp, 1] = raw_to_volts(l0["PA_LFR_SC_E1_" + freq][packet, :])
        data["v"][i:i + nsamp, 2] = raw_to_volts(l0["PA_LFR_SC_E2_" + freq][packet, :])

        i += nsamp

    return data

def raw_to_volts(raw_value):
    """
    Convert LFR V raw value to Volts units.

    Comment from Bias team:
    'for DC single(which we have for mux=4, BIAS_1-3), and BIAS+LFR,
    the conversion should be _approximately_ V_TM = V_Volt * 8000 * 1/17'

    :param raw_value: LFR F3 V value in count. Can be a scalar or a numpy array
    :return: value in Volts
    """
    return raw_value * BIA_SWEEP_TM2V_FACTOR


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import namedtuple

import numpy as np
from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records

from roc.rap.tasks.utils import sort_data
from roc.rap.tasks.lfr.utils import set_cwf
from roc.rap.tasks.lfr.bp import convert_BP as bp_lib

__all__ = ["bp1", "bp2", "cwf"]


class L1LFRSBM2Error(Exception):
    """
    Errors for L1 SBM2 LFR.
    """


# Numpy array dtype for LFR BP1 L1 SBM2 data
# Name convention is lower case
BP1 = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", "uint32", 2),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("bias_mode_mux_set", "uint8"),
    ("bias_mode_hv_enabled", "uint8"),
    ("bias_mode_bias1_enabled", "uint8"),
    ("bias_mode_bias2_enabled", "uint8"),
    ("bias_mode_bias3_enabled", "uint8"),
    ("bias_on_off", "uint8"),
    ("bw", "uint8"),
    ("sp0", "uint8"),
    ("sp1", "uint8"),
    ("r0", "uint8"),
    ("r1", "uint8"),
    ("r2", "uint8"),
    ("pe", "float64", 26),
    ("pb", "float64", 26),
    ("nvec_v0", "float32", 26),
    ("nvec_v1", "float32", 26),
    ("nvec_v2", "uint8", 26),
    ("ellip", "float32", 26),
    ("dop", "float32", 26),
    ("sx_rea", "float64", 26),
    ("sx_arg", "uint8", 26),
    ("vphi_rea", "float64", 26),
    ("vphi_arg", "uint8", 26),
    ("freq", "uint8"),
]


def set_bp1(fdata, freq):

    # Sort input packets by ascending acquisition times
    fdata, __ = sort_data(fdata, (
        fdata["PA_LFR_ACQUISITION_TIME"][:, 1],
        fdata["PA_LFR_ACQUISITION_TIME"][:, 0],
    ))

    # create a record array with the good dtype and good length
    result = np.empty(
        fdata["PA_LFR_ACQUISITION_TIME"].shape[0],
        dtype=BP1,
    )
    fill_records(result)

    # get the shape of the data
    shape = fdata["PA_LFR_SC_BP1_PE_" + freq].shape

    # Get number of packets and blocks
    n = shape[1]

    # get the acquisition time and other data
    result["acquisition_time"] = fdata["PA_LFR_ACQUISITION_TIME"][:, :2]
    result["synchro_flag"] = fdata["PA_LFR_ACQUISITION_TIME"][:, 2]
    result["bias_mode_mux_set"] = fdata["PA_BIA_MODE_MUX_SET"]
    result["bias_mode_hv_enabled"] = fdata["PA_BIA_MODE_HV_ENABLED"]
    result["bias_mode_bias1_enabled"] = fdata[
        "PA_BIA_MODE_BIAS1_ENABLED"
    ]
    result["bias_mode_bias2_enabled"] = fdata[
        "PA_BIA_MODE_BIAS2_ENABLED"
    ]
    result["bias_mode_bias3_enabled"] = fdata[
        "PA_BIA_MODE_BIAS3_ENABLED"
    ]
    result["bias_on_off"] = fdata["PA_BIA_ON_OFF"]
    result["bw"] = fdata["SY_LFR_BW"]
    result["sp0"] = fdata["SY_LFR_SP0"]
    result["sp1"] = fdata["SY_LFR_SP1"]
    result["r0"] = fdata["SY_LFR_R0"]
    result["r1"] = fdata["SY_LFR_R1"]
    result["r2"] = fdata["SY_LFR_R2"]

    # Convert BP1 values (see convert_BP.py)
    struct = namedtuple("struct", ["nbitexp", "nbitsig", "rangesig_16",
                                   "rangesig_14", "expmax", "expmin"])
    bp_lib.init_struct(struct)

    result['pe'][:, :n] = bp_lib.convToFloat16(struct,
                                               fdata["PA_LFR_SC_BP1_PE_" + freq][:, :n])
    result['pb'][:, :n] = bp_lib.convToFloat16(struct,
                                               fdata["PA_LFR_SC_BP1_PB_" + freq][:, :n])
    # Compute n1 component of k-vector (normalized)
    result['nvec_v0'][:, :n] = bp_lib.convToFloat8(
        fdata["PA_LFR_SC_BP1_NVEC_V0_" + freq][:, :n])
    # Compute n2 component of k-vector (normalized)
    result['nvec_v1'][:, :n] = bp_lib.convToFloat8(
        fdata["PA_LFR_SC_BP1_NVEC_V1_" + freq][:, :n])

    # Store the 1b bit of the sign of n3 components k-vector in nvec_v2
    result['nvec_v2'][:, :n] = bp_lib.convToUint8(
        fdata["PA_LFR_SC_BP1_NVEC_" + freq][:, :n])

    result['ellip'][:, :n] = bp_lib.convToFloat4(
        fdata["PA_LFR_SC_BP1_ELLIP_" + freq][:, :n])

    result['dop'][:, :n] = bp_lib.convToFloat3(
        fdata["PA_LFR_SC_BP1_DOP_" + freq][:, :n])

    # Extract x-component of normalized Poynting flux...
    # sign (1 bit)...
    sx_sign = 1 - \
        (2 * bp_lib.Uint16ToUint8(fdata["PA_LFR_SC_BP1_SX_" +
                                        freq][:, :n], 15, 0x1).astype(np.float))

    # real value (14 bits)
    result['sx_rea'][:, :n] = sx_sign * bp_lib.convToFloat14(struct,
                                                             fdata["PA_LFR_SC_BP1_SX_" + freq][:, :n])

    # and arg (1 bit)
    result['sx_arg'][:, :n] = bp_lib.Uint16ToUint8(
        fdata["PA_LFR_SC_BP1_SX_" + freq][:, :n], 14, 0x1)

    # Extract velocity phase...

    # sign (1 bit)...
    vphi_sign = 1 - (2 * bp_lib.Uint16ToUint8(
        fdata["PA_LFR_SC_BP1_VPHI_" + freq][:, :n], 15, 0x1).astype(np.float))

    # real value (14 bits)
    result['vphi_rea'][:, :n] = vphi_sign * bp_lib.convToFloat14(struct,
                                                                 fdata["PA_LFR_SC_BP1_VPHI_" + freq][:, :n])

    # and arg (1 bit)
    result['vphi_arg'][:, :n] = bp_lib.Uint16ToUint8(
        fdata["PA_LFR_SC_BP1_VPHI_" + freq][:, :n], 14, 0x1)

    # Get TT2000 Epoch time
    result["epoch"] = Time().obt_to_utc(
        result["acquisition_time"], to_tt2000=True)

    # Get scet
    result["scet"] = Time.cuc_to_scet(result["acquisition_time"])

    return result


def bp1(f, task):
    def get_data(freq):
        # get name of the packet
        name = "TM_LFR_SCIENCE_SBM2_BP1_F" + freq
        logger.debug("Get data for " + name)

        # get the sbm2 bp1 data
        if name in f["TM"]:
            # get data from the file
            fdata = f["TM"][name]["source_data"]

            # get data
            data = set_bp1(fdata, "F" + freq)

            data["freq"][:] = int(freq)

        else:
            data = np.empty(0, dtype=BP1)
        return data

    # get the data of all packets
    data = [get_data(x) for x in "01"]

    # concatenate both arrays
    return np.concatenate(tuple(data))


# Numpy array dtype for LFR BP2 L1 SBM2 data
# Name convention is lower case
BP2 = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", "uint32", 2),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("bias_mode_mux_set", "uint8"),
    ("bias_mode_hv_enabled", "uint8"),
    ("bias_mode_bias1_enabled", "uint8"),
    ("bias_mode_bias2_enabled", "uint8"),
    ("bias_mode_bias3_enabled", "uint8"),
    ("bias_on_off", "uint8"),
    ("bw", "uint8"),
    ("sp0", "uint8"),
    ("sp1", "uint8"),
    ("r0", "uint8"),
    ("r1", "uint8"),
    ("r2", "uint8"),
    ("auto", "float64", (26, 5)),
    ("cross_re", "float32", (26, 10)),
    ("cross_im", "float32", (26, 10)),
    ("freq", "uint8"),
]


def set_bp2(fdata, freq):

    # Sort input packets by ascending acquisition times
    fdata, __ = sort_data(fdata, (
        fdata["PA_LFR_ACQUISITION_TIME"][:, 1],
        fdata["PA_LFR_ACQUISITION_TIME"][:, 0],
    ))

    # create a record array with the good dtype and good length
    result = np.empty(
        fdata["PA_LFR_ACQUISITION_TIME"].shape[0],
        dtype=BP2,
    )
    fill_records(result)

    # get shape of data
    shape = fdata["PA_LFR_SC_BP2_AUTO_A0_" + freq].shape

    # get the acquisition time and other data
    result["acquisition_time"] = fdata["PA_LFR_ACQUISITION_TIME"][:, :2]
    result["synchro_flag"] = fdata["PA_LFR_ACQUISITION_TIME"][:, 2]
    result["bias_mode_mux_set"] = fdata["PA_BIA_MODE_MUX_SET"]
    result["bias_mode_hv_enabled"] = fdata["PA_BIA_MODE_HV_ENABLED"]
    result["bias_mode_bias1_enabled"] = fdata[
        "PA_BIA_MODE_BIAS1_ENABLED"
    ]
    result["bias_mode_bias2_enabled"] = fdata[
        "PA_BIA_MODE_BIAS2_ENABLED"
    ]
    result["bias_mode_bias3_enabled"] = fdata[
        "PA_BIA_MODE_BIAS3_ENABLED"
    ]
    result["bias_on_off"] = fdata["PA_BIA_ON_OFF"]
    result["bw"] = fdata["SY_LFR_BW"]
    result["sp0"] = fdata["SY_LFR_SP0"]
    result["sp1"] = fdata["SY_LFR_SP1"]
    result["r0"] = fdata["SY_LFR_R0"]
    result["r1"] = fdata["SY_LFR_R1"]
    result["r2"] = fdata["SY_LFR_R2"]

    auto = np.dstack(
        list(
            fdata["PA_LFR_SC_BP2_AUTO_A{0}_{1}".format(x, freq)]
            for x in range(5)
        )
    )
    cross_re = np.dstack(
        list(
            fdata["PA_LFR_SC_BP2_CROSS_RE_{0}_{1}".format(x, freq)]
            for x in range(10)
        )
    )
    cross_im = np.dstack(
        list(
            fdata["PA_LFR_SC_BP2_CROSS_IM_{0}_{1}".format(x, freq)]
            for x in range(10)
        )
    )

    # Get number of blocks
    n = shape[1]

    # Convert AUTO, CROSS_RE an CROSS_IM (see convert_BP.py)
    struct = namedtuple("struct", ["nbitexp", "nbitsig", "rangesig_16", "rangesig_14",
                                   "expmax", "expmin"])
    bp_lib.init_struct(struct)

    result['auto'][:, :n, :] = bp_lib.convToAutoFloat(struct, auto)
    result['cross_re'][:, :n, :] = bp_lib.convToCrossFloat(cross_re)
    result['cross_im'][:, :n, :] = bp_lib.convToCrossFloat(cross_im)

    # WARNING, IF NAN or INF, put FILLVAL
    # TODO - Optimize this part
    result['auto'][np.isinf(result['auto'])] = -1e31
    result['auto'][np.isnan(result['auto'])] = -1e31
    result['cross_re'][np.isinf(result['cross_re'])] = -1e31
    result['cross_re'][np.isnan(result['cross_re'])] = -1e31
    result['cross_im'][np.isinf(result['cross_im'])] = -1e31
    result['cross_im'][np.isnan(result['cross_im'])] = -1e31

    # Get TT2000 Epoch times
    result["epoch"] = Time().obt_to_utc(
        result["acquisition_time"], to_tt2000=True)

    # Get scet
    result["scet"] = Time.cuc_to_scet(result["acquisition_time"])

    return result


def bp2(f, task):
    def get_data(freq):
        # get name of the packet
        name = "TM_LFR_SCIENCE_SBM2_BP2_F" + freq
        logger.debug("Get data for " + name)

        # get the sbm2 bp1 data
        if name in f["TM"]:
            # get data from the file
            fdata = f["TM"][name]["source_data"]

            #  get the data
            data = set_bp2(fdata, "F" + freq)

            data["freq"][:] = int(freq)

        else:
            data = np.empty(0, dtype=BP2)

        return data

    # get data for all packets
    data = [get_data(x) for x in "01"]

    # concatenate both arrays
    return np.concatenate(tuple(data))

# Numpy array dtype for LFR CWF L1 SBM2 data
# Name convention is lower case
CWF = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", "uint32", 2),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("bias_mode_mux_set", "uint8"),
    ("bias_mode_hv_enabled", "uint8"),
    ("bias_mode_bias1_enabled", "uint8"),
    ("bias_mode_bias2_enabled", "uint8"),
    ("bias_mode_bias3_enabled", "uint8"),
    ("bias_on_off", "uint8"),
    ("bw", "uint8"),
    ("sp0", "uint8"),
    ("sp1", "uint8"),
    ("r0", "uint8"),
    ("r1", "uint8"),
    ("r2", "uint8"),
    ("sampling_rate", "float32"),
    ("v", "int16"),
    ("e", "int16", 2),
    ("b", "int16", 3),
]


def cwf(f, task):
    # get name of the packet
    name = "TM_LFR_SCIENCE_SBM2_CWF_F2"
    logger.debug("Get data for " + name)

    # get the sbm1 bp1 data
    if name in f["TM"]:
        fdata = f["TM"][name]["source_data"]
    else:
        return np.empty(
            0,
            dtype=CWF,
        )

    ddata = {
        name: fdata[name][...] for name in fdata.keys()
    }

    # get the data for cwf
    return set_cwf(ddata, "F2", CWF, True, task)



import numpy as np
cimport numpy as np

from roc.rpl.packet_structure.data import Data
from roc.rpl.packet_structure.data cimport Data

cpdef unsigned long compute_tempo(unsigned long frequency):
    """
    Return the time for measurement of the frequency in millisecond.
    """
    cdef:
        unsigned long tempo

    tempo = 11
    if frequency > 16000:
        tempo += 19
        return tempo
    if frequency > 12000:
        tempo += 11
        return tempo
    if frequency > 7000:
        tempo += 6
        return tempo
    return tempo

cpdef unsigned short hfr_band_index(str band):
    cdef:
        unsigned short band_index
    if band == "hf1":
        band_index = 1
    elif band == "hf2":
        band_index = 2
    else:
        raise Exception("Invalid HFR band name: {0}".format(band))

    return band_index

cpdef unsigned long compute_frequency(
    unsigned short init_frequency,
    unsigned short step,
    unsigned long hf1_step,
    unsigned long hf1_number,
    unsigned long hf2_step,
    str band,
):
    cdef:
        unsigned long number

    # compute the start frequency
    number = 436 + init_frequency

    # hf1 band
    if band == "hf1":
        number += step * hf1_step

    # hf2 band
    if band == "hf2":
        number += (hf1_number - 1) * hf1_step
        number += (step + 1) * hf2_step

    # compute the frequency in kHz from the formula given in the TNR/HFR
    # software user manual
    return 375 + 50 * (number - 436)


cpdef list hfr_dtype():
    """
    Numpy array dtype for the HFR L1 survey data.
    """
    # Name convention is lower case
    return [
        ("epoch", "float64"),
        ("scet", "float64"),
        ("acquisition_time", "uint32", 2),
        ("synchro_flag", "uint8"),
        ("quality_flag", "uint8"),
        ("quality_bitmask", "uint8"),
        ("sweep_num", "uint32"),
        ("sample_time", "float64"),
        ("ticks_nr", "int64"),
        ("delta_time", "float64"),
        ("sweep_mode", "uint8"),
        ("survey_mode", "uint8"),
        ("channel_status", "uint8", 2),
        ("calibration_level", "uint8"),
        ("average_nr", "uint8"),
        ("front_end", "uint8"),
        ("sensor_config", "uint8", 2),
        ("rpw_status", "uint8", 15),
        ("temperature", "uint8", 4),
        ("hfr_band", "uint8"),
        ("frequency", "uint16"),
        ("agc1", "uint16"),
        ("agc2", "uint16"),
    ]


cpdef tuple delta_times(Data data, unsigned int start):
    """
    Read delta times command.
    """
    cdef:
        long long hf1
        long long hf2

    hf1 = <long long>data.u32(start, 0, 32)
    hf2 = <long long>data.u32(start + 4, 0, 32)

    return hf1, hf2


cpdef tuple hfr_sweep_setup (Data data, unsigned int start):
    """
    Return tuple of values from the HFR sweep setup command.
    """
    cdef:
        unsigned char hf1_size
        unsigned short hf1_number
        unsigned char hf2_size
        unsigned short hf2_number

    # read parameters
    hf1_size = data.u8(start + 3, 4, 4)
    hf1_number = data.u16(start + 2, 3, 9)
    hf2_size = data.u8(start + 1, 7, 4)
    hf2_number = data.u16(start, 6, 9)

    return hf2_number, hf2_size, hf1_number, hf1_size


cpdef tuple hfr_setup(Data data, unsigned int start):
    """
    Return tuple of values from the HFR setup command.
    """
    cdef:
        unsigned char av
        unsigned char hf1
        unsigned char hf2
        unsigned short initial_frequency
        unsigned char sw
        unsigned int tmp

    tmp = start + 1

    # read parameters
    av = data.u8(tmp, 6, 2)
    hf1 = data.u8(tmp, 5, 1)
    hf2 = data.u8(tmp, 4, 1)
    initial_frequency = data.u16(start, 3, 9)
    sw = data.u8(start, 2, 1)

    return sw, initial_frequency, hf2, hf1, av

cpdef tuple extract_band(
    unsigned char ch,
    unsigned char band1,
    unsigned char band2,
    unsigned short n1,
    unsigned short n2,
    Data data,
    unsigned int offset,
    unsigned int bit_count,
):
    """
    Extract the band, returning fake data if not selected.
    """
    cdef:
        Py_ssize_t i
        unsigned short ch1_value
        unsigned short ch2_value
        np.ndarray[np.uint16_t, ndim=1] band1_value = np.zeros(512, dtype="uint16")
        np.ndarray[np.uint16_t, ndim=1] band2_value = np.zeros(512, dtype="uint16")

    # if channel is activated
    if ch == 1:
        # read the first band
        if band1 == 1:
            for i in range(n1):
                band1_value[i] = data.u16(
                    offset + bit_count // 8,
                    bit_count % 8,
                    12,
                )
                bit_count += 12

        # second band
        if band2 == 1:
            for i in range(n2):
                band2_value[i] = data.u16(
                    offset + bit_count // 8,
                    bit_count % 8,
                    12,
                )
                bit_count += 12

    return (
        (
            band1_value, band2_value
        ), bit_count
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from datetime import datetime

from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.rpl.packet_parser import raw_to_eng
from roc.rpl.constants import TIME_ISO_STRFORMAT

# Numpy array dtype for BIAS current data
# Name convention is lower case
from roc.rap import MetadataException

CURRENT_DTYPE = [
    ("epoch", "uint64"),
    ("ibias_1", "float32"),
    ("ibias_2", "float32"),
    ("ibias_3", "float32"),
]

# Number of antennas (3)
ANT_NUM = 3

# SRDB ID of the transfer function TF_CP_BIA_P011 (or TF_CP_BIA_0011 in ICD)
# NOTES: SRDB_ID is CIWP0075TC and not CIWP0040TC as indicated in the ICD)
TF_CP_BIA_P011_SRDB_ID = 'CIWP0075TC'

def set_current(l0, task):
    """
    Set Bias current values from TC_DPU_SET_BIAi

    :param l0: input l0 files
    :return:
    """

    from roc.idb.converters.cdf_fill import fill_records

    # Get/create time instance
    time = Time()

    # Get number of TC_DPU_SET_BIAS1/2/3 packets in L0
    # (One packet == one record)
    sizes = np.zeros(ANT_NUM, dtype=np.uint32)
    for i in range(ANT_NUM):
        tc_name = 'TC_DPU_SET_BIAS{0}'.format(i + 1)
        if tc_name in l0['TC']:
            ack_exe_state = l0['TC'][tc_name]['tc_ack_state'][:,1]
            where_exe_passed = ack_exe_state[ack_exe_state == 'PASSED']
            sizes[i] = where_exe_passed.shape[0]

    # Get number of output records (one by TC_DPU_SET_BIAS1/2/3 packets)
    nrec = np.sum(sizes)
    # And initialize the output numpy array for BIAS current L1 data
    data = np.empty(nrec, dtype=CURRENT_DTYPE)
    if nrec > 0:
        # create a record array with the good dtype and good length
        fill_records(data)
    else:
        # if no data, return empty array
        return data

    # Fill the data numpy array
    counter = 0
    for i in range(ANT_NUM):
        tc_name = 'TC_DPU_SET_BIAS{0}'.format(i + 1)
        if tc_name not in l0['TC']:
            continue
        else:
            ack_exe_state = l0['TC'][tc_name]['tc_ack_state'][:,1]
            where_exe_passed = ack_exe_state == 'PASSED'

        # If not 'passed' TC, then continue
        if not where_exe_passed.any():
            continue
        else:
            ibias = l0['TC'][tc_name]['application_data']['CP_BIA_SET_BIAS{0}'.format(i + 1)][where_exe_passed]
            utc_time = l0['TC'][tc_name]['utc_time'][where_exe_passed]
            idb_source = l0['TC'][tc_name]['idb'][where_exe_passed, 0]
            idb_version = l0['TC'][tc_name]['idb'][where_exe_passed, 1]
            size = len(ibias)

        # Else store the bias current intensity values found in the output sample array
        # TODO - See how to deal if IDB version change in the output sample array
        data['ibias_{0}'.format(i+1)][counter:counter + size] = raw_to_na(ibias,
                                                                          idb_source=idb_source[0],
                                                                          idb_version=idb_version[0])

        # And corresponding TC execution time (UTC in input L0)
        data['epoch'][counter:counter + size] = [time.utc_to_tt2000(datetime.strptime(current_time[:-4],
                                                                                      TIME_ISO_STRFORMAT[:-1]))
                                                 for current_time in utc_time]

        # Increment counter
        counter += size

    return data

def raw_to_na(raw_values,
              idb_source="MIB",
              idb_version=None):
    """
    Convert input raw values of bias current into physical units (nA)

    :param raw_values: numpy array with raw values of Bias current
    :param idb_version: string with idb version
    :param idb_source: string with idb_source
    :return: values in physical units (nA)
    """

    # Retrieve engineering values in uA and return them in nA
    return raw_to_eng(raw_values, TF_CP_BIA_P011_SRDB_ID,
                    idb_source=idb_source, idb_version=idb_version) * 1000



import numpy as np
cimport numpy as np

from libc.stdint cimport uint8_t
from libc.stdint cimport uint16_t
from libc.stdint cimport uint32_t
from libc.stdint cimport uint64_t

from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records

from roc.rap.tasks.tds.utils import group

from roc.rap.tasks.utils import sort_data

__all__ = ["HIST2D", "set_hist2d"]

# Numpy array dtype for TDS histo2D L1 survey data
# Name convention is lower case

HIST2D = [
    ("epoch", "uint64"),
    ("scet", "float64"),
    ("acquisition_time", ("uint32", 2)),
    ("synchro_flag", "uint8"),
    ("quality_flag", "uint8"),
    ("quality_bitmask", "uint8"),
    ("survey_mode", "uint8"),
    ("bia_status_info", ("uint8", 6)),
    ("channel_status_info", ("uint8", 4)),
    ("sampling_rate", "float32"),
    ("input_config", "uint32"),
    ("snapshot_len", "uint8"),
    ("hist2d_id", "uint8"),
    ("hist2d_params", "uint8"),
    ("hist2d_axis1", "uint8"),
    ("hist2d_axis2", "uint8"),
    ("hist2d_col_time", "uint16"),
    ("hist2d_bins1", "uint8"),
    ("hist2d_bins2", "uint8"),
    ("hist2d_tot_pts", "uint16"),
    ("hist2d_counts", ("uint16", (128, 128))),
]

# TDS HIST2D sampling rate in Hz
SAMPLING_RATE_HZ = [65534.375, 131068.75, 262137.5, 524275., 2097100.]

cpdef set_hist2d(l0, task):
    cdef:
        uint64_t size
        uint64_t index
        uint64_t packet
        uint64_t grp
        uint16_t data_space
        uint16_t data_space_current
        uint16_t matrix_size
        uint16_t nx, ny

    # name of the packet
    name = "TM_TDS_SCIENCE_NORMAL_HIST2D"
    logger.debug("Get data for " + name)

    # get the HIST2D packet data
    if name in l0["TM"]:
        tm = l0["TM"][name]["source_data"]
    else:
        # rest of the code inside the with statement will not be called
        return np.empty(
            0,
            dtype=HIST2D,
        )

    # Sort input packets by ascending acquisition times and hist2D group number
    tm, __ = sort_data(tm, (
        tm["PA_TDS_ACQUISITION_TIME"][:,1],
        tm["PA_TDS_ACQUISITION_TIME"][:,0],
                    ))

    # get the size of the data from the number of groups
    groups, size = group(tm["PA_TDS_HIST2D_PKT_NR"][:])

    # create the data array with good size
    data = np.empty(size, dtype=HIST2D)
    fill_records(data)

    # get the number of packets
    size = tm["PA_TDS_ACQUISITION_TIME"].shape[0]

    # fake current group for start
    grp = -1

    # loop over packets
    with nogil:
        while (packet < size):
            with gil:
                # if the group as changed
                if groups[packet] != grp:
                    # change the current group
                    grp = groups[packet]

                    # make the initialization of the data

                    # Time variable
                    data["epoch"][grp] = Time().obt_to_utc(tm[
                        "PA_TDS_ACQUISITION_TIME"
                    ][packet, :2].reshape([1, 2]), to_tt2000=True)
                    data["acquisition_time"][grp, :] = tm[
                        "PA_TDS_ACQUISITION_TIME"
                    ][packet, :2]
                    data["synchro_flag"][grp] = tm["PA_TDS_ACQUISITION_TIME"][
                        packet,
                        2,
                    ]

                    # scet
                    data["scet"][grp] = Time.cuc_to_scet(data["acquisition_time"][grp, :])


                    # bias info
                    data["bia_status_info"][grp, :] = (
                        tm["PA_BIA_MODE_MUX_SET"][packet],
                        tm["PA_BIA_MODE_HV_ENABLED"][packet],
                        tm["PA_BIA_MODE_BIAS1_ENABLED"][packet],
                        tm["PA_BIA_MODE_BIAS2_ENABLED"][packet],
                        tm["PA_BIA_MODE_BIAS3_ENABLED"][packet],
                        tm["PA_BIA_ON_OFF"][packet],
                        )

                    data["channel_status_info"][grp, :] = (
                        tm["PA_TDS_STAT_ADC_CH1"][packet],
                        tm["PA_TDS_STAT_ADC_CH2"][packet],
                        tm["PA_TDS_STAT_ADC_CH3"][packet],
                        tm["PA_TDS_STAT_ADC_CH4"][packet],
                    )

                    # Survey mode (always normal here == 3)
                    data["survey_mode"][grp] = 3

                    # sampling rate
                    data["sampling_rate"][grp] = SAMPLING_RATE_HZ[tm["PA_TDS_STAT_SAMP_RATE"][packet]]

                    # input config
                    data["input_config"][grp] = tm["PA_TDS_SWF_HF_CH_CONF"][packet]

                    # hist parameters
                    data["snapshot_len"] = tm["PA_TDS_SNAPSHOT_LEN"][packet]
                    data["hist2d_id"][grp] = tm["PA_TDS_SC_HIST2D_ID"][packet]
                    data["hist2d_params"][grp] = tm["PA_TDS_SC_HIST2D_PARAMS"][packet]
                    data["hist2d_axis1"][grp] = tm["PA_TDS_SC_HIST2D_PAR1_AXIS"][packet]
                    data["hist2d_axis2"][grp] = tm["PA_TDS_SC_HIST2D_PAR2_AXIS"][packet]
                    data["hist2d_bins1"][grp] = tm["PA_TDS_SC_HIST2D_PAR1_BINS"][packet]
                    data["hist2d_bins2"][grp] = tm["PA_TDS_SC_HIST2D_PAR2_BINS"][packet]
                    data["hist2d_col_time"][grp] = tm[
                        "PA_TDS_SC_HIST2D_COLL_TIME"
                    ][packet]
                    data["hist2d_tot_pts"][grp] = tm[
                        "PA_TDS_SC_HIST2D_TOT_POINTS"
                    ][packet]

                    # size of the matrix for counts
                    nx = data["hist2d_bins1"][grp]
                    ny = data["hist2d_bins2"][grp]
                    matrix_size = nx * ny
                    logger.debug(
                        "Matrix size for packet {0} is {1}".format(
                            packet,
                            matrix_size,
                        )
                    )

                    # create the memory space for the count matrix
                    matrix = np.zeros(matrix_size, dtype="uint16")

                    # compute the space for the data
                    if (
                        tm["PA_TDS_HIST2D_PKT_NR"][packet] ==
                        (tm["PA_TDS_HIST2D_PKT_CNT"][packet] - 1)
                    ):
                        # this is the last packet, all other are missing so the size of
                        # the data space is the one of the matrix minius the size of
                        # the data of this packet
                        data_space = (
                            matrix_size - tm["PA_TDS_SC_HIST2D_DATA_NR"][packet]
                        ) // tm["PA_TDS_HIST2D_PKT_CNT"][packet]
                    else:
                        # the size can be read directly from the packet because all
                        # except the last one have the same size
                        data_space = tm["PA_TDS_SC_HIST2D_DATA_NR"][packet]
                    logger.debug(
                        "Data space from packet number {0} is {1}".format(
                            packet,
                            data_space,
                        )
                    )

                # for each packet add the data
                # get the index in the snapshot of the data
                index = tm["PA_TDS_HIST2D_PKT_NR"][packet] - 1

                # get the space taken on the matrix for this data
                data_space_current = tm["PA_TDS_SC_HIST2D_DATA_NR"][packet]
                logger.debug(
                    "Packet index in snapshot: {0} for packet {1} & chunk {2}".format(
                        index,
                        packet,
                        data_space,
                    )
                )

                # fill the matrix in part with the data
                matrix[
                    index * data_space: index * data_space + data_space_current
                ] = tm["PA_TDS_SC_HIST2D_DATA"][packet][:data_space_current]

                # if last packet, copy the data to the structure
                if index == (tm["PA_TDS_HIST2D_PKT_CNT"][packet] - 1):
                    data["hist2d_counts"][grp][:nx, :ny] = matrix.reshape(
                        (nx, ny)
                    )[:, :]

                # increment the packet
                packet += 1



    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
cimport numpy as np


__all__ = ["find", "group", "count_swf"]


def find(iterable, name):
    """
    Find the parameter with name in the list in argument.
    """
    for x in iterable:
        if x.name == name:
            return x

    raise ValueError("Parameter {0} not found in list of parameters")


cpdef tuple group(np.ndarray[np.uint8_t, ndim=1] x):
    """
    Method use to group HIST2D packets (with PA_TDS_HIST2D_PKT_NR)
    
    :param x: 
    :return: 
    """
    cdef:
        Py_ssize_t i
        Py_ssize_t nmax = x.shape[0]
        np.ndarray result = np.zeros(nmax, dtype=np.uint64)
        Py_ssize_t counter = 0

    for i in range(1, nmax):
        if x[i] <= x[i-1]:
            counter += 1
        result[i] = counter

    return result, counter + 1


cpdef tuple count_swf(np.ndarray[np.uint32_t, ndim=1] x):
    """
    Method use to count SWF packets using index of snapshots.
    
    :param x: snapshot ids in packets
    :return: 2-elements tuple with corresponding list of counters and counter max. value
    """
    cdef:
        Py_ssize_t i
        Py_ssize_t nmax = x.shape[0]
        np.ndarray result = np.zeros(nmax, dtype=np.uint32)
        Py_ssize_t counter = 0

    for i in range(1, nmax):
        if x[i] != x[i-1]:
            counter += 1
        result[i] = counter

    return result, counter + 1


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import h5py
from poppy.core.logger import logger


__all__ = ['PLUGIN',
           'order_by_increasing_time',
           'sort_data',
           'h5_group_to_dict']



# the name of the plugin
PLUGIN = 'roc.rap'

def order_by_increasing_time(data,
                             sort_by='epoch',
                             unique=False):
    """
    Given a structured array of data containing an epoch or acquisition_time field, the
    data is reordered by its values, and returns a new structured array of the
    ordered data.
    If no Epoch found then acquisition time is used instead.

    :param data: Input numpy.ndarray object to sort by ascending time
    :param sort_by: Name of the time variable used to sort data ("epoch" by defaut)
    :param unique; If True, remove duplicate Epoch data
    :return: sorted numpy.ndarray object
    """

    logger.debug('Ordering data by increasing time')
    if sort_by.lower() == 'epoch' and 'epoch' in data.dtype.names:
        if unique:
            _, indices = np.unique(data['epoch'], return_index=True)
            data = data[indices[:]]

        # If epoch found, then sort by ascending epoch values (nanosec resolution)
        data.sort(order='epoch')

    elif sort_by.lower() == 'acquisition_time' and 'acquisition_time' in data.dtype.names:
        # If not epoch found, then use CCSDS CUC acquisition time (~15 microsec resolution)
        if unique:
            _, indices = np.unique(data['acquisition_time'], return_index=True, axis=0)
            data = data[indices[:]]

        # get the indices of the reordering (with lexsort, the last column in the
        # array is the primary column for the sort).
        indices = np.lexsort(
            (
            data['acquisition_time'][:,1],
            data['acquisition_time'][:,0],
            )
        )
        data = data[indices[:]]
    else:
        logger.debug('No valid time variable, skip time ordering')

    # return the ordered data
    return data


def h5_group_to_dict(h5_group):
    """
    Return HDF5 group as a dictionary

    :param h5_group: HDF5 subgroup
    :return: dictionary
    """
    return {
        name: h5_group[name][()]
        for name in h5_group.keys()
    }

def sort_data(data, sort_by):
    """
    Sort input data using numpy.lexsort method().

    :param data: dictionary to sort
    :param sort_by: tuple containing data variables to sort by
    :return: sorted data
    """

    # If input data is a h5py group then
    # copy h5 data into a dictionary (to avoid change in the h5 file)
    if isinstance(data, h5py.Group):
        data = h5_group_to_dict(data)

    # Get sorted indices
    indices = np.lexsort(sort_by)
    for name in data.keys():
        data[name] = data[name][indices[:]]

    # Return sorted data and related indices
    return data, indices



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import namedtuple

import numpy as np
from poppy.core.logger import logger

from roc.rpl.time import Time
from roc.idb.converters.cdf_fill import fill_records

from roc.rap.tasks.utils import sort_data
from roc.rap.tasks.lfr.utils import set_cwf
from roc.rap.tasks.lfr.bp import convert_BP as bp_lib

__all__ = ['bp1', 'bp2', 'cwf']


class L1LFRSBM1Error(Exception):
    """
    Errors for L1 SBM1 LFR.
    """


# Numpy array dtype for LFR BP1 L1 SBM1 data
# Name convention is lower case
BP1 = [
    ('epoch', 'uint64'),
    ('scet', 'float64'),
    ('acquisition_time', 'uint32', 2),
    ('synchro_flag', 'uint8'),
    ('quality_flag', 'uint8'),
    ('quality_bitmask', 'uint8'),
    ('bias_mode_mux_set', 'uint8'),
    ('bias_mode_hv_enabled', 'uint8'),
    ('bias_mode_bias1_enabled', 'uint8'),
    ('bias_mode_bias2_enabled', 'uint8'),
    ('bias_mode_bias3_enabled', 'uint8'),
    ('bias_on_off', 'uint8'),
    ('bw', 'uint8'),
    ('sp0', 'uint8'),
    ('sp1', 'uint8'),
    ('r0', 'uint8'),
    ('r1', 'uint8'),
    ('r2', 'uint8'),
    ('pe', 'float64', 22),
    ('pb', 'float64', 22),
    ('nvec_v0', 'float32', 22),
    ('nvec_v1', 'float32', 22),
    ('nvec_v2', 'uint8', 22),
    ('ellip', 'float32', 22),
    ('dop', 'float32', 22),
    ('sx_rea', 'float64', 22),
    ('sx_arg', 'uint8', 22),
    ('vphi_rea', 'float64', 22),
    ('vphi_arg', 'uint8', 22),
]


def bp1(f, task):
    # name of the packet
    freq = 'F0'
    name = 'TM_LFR_SCIENCE_SBM1_BP1_' + freq
    logger.debug('Get data for ' + name)

    # get the sbm1 bp1 data
    if name in f['TM']:
        fdata = f['TM'][name]['source_data']
    else:
        # rest of the code inside the with statement will not be called
        return np.empty(
            0,
            dtype=BP1,
        )

    # Sort input packets by ascending acquisition times
    fdata, __ = sort_data(fdata, (
        fdata['PA_LFR_ACQUISITION_TIME'][:, 1],
        fdata['PA_LFR_ACQUISITION_TIME'][:, 0],
    ))

    # create an array for the data
    result = np.empty(
        fdata['PA_LFR_ACQUISITION_TIME'].shape[0],
        dtype=BP1,
    )
    fill_records(result)

    # Get number of packets and blocks
    shape = fdata['PA_LFR_SC_BP1_PE_' + freq].shape
    n = shape[1]

    # get the acquisition time and other data
    result['acquisition_time'] = fdata['PA_LFR_ACQUISITION_TIME'][:, :2]
    result['synchro_flag'] = fdata['PA_LFR_ACQUISITION_TIME'][:, 2]
    result['bias_mode_mux_set'] = fdata['PA_BIA_MODE_MUX_SET']
    result['bias_mode_hv_enabled'] = fdata['PA_BIA_MODE_HV_ENABLED']
    result['bias_mode_bias1_enabled'] = fdata['PA_BIA_MODE_BIAS1_ENABLED']
    result['bias_mode_bias2_enabled'] = fdata['PA_BIA_MODE_BIAS2_ENABLED']
    result['bias_mode_bias3_enabled'] = fdata['PA_BIA_MODE_BIAS3_ENABLED']
    result['bias_on_off'] = fdata['PA_BIA_ON_OFF']
    result['bw'] = fdata['SY_LFR_BW']
    result['sp0'] = fdata['SY_LFR_SP0']
    result['sp1'] = fdata['SY_LFR_SP1']
    result['r0'] = fdata['SY_LFR_R0']
    result['r1'] = fdata['SY_LFR_R1']
    result['r2'] = fdata['SY_LFR_R2']

    # Convert BP1 values (see convert_BP.py)
    struct = namedtuple('struct', ['nbitexp', 'nbitsig', 'rangesig_16',
                                   'rangesig_14', 'expmax', 'expmin'])
    bp_lib.init_struct(struct)

    result['pe'][:, :n] = bp_lib.convToFloat16(struct,
                                               fdata['PA_LFR_SC_BP1_PE_' + freq][:, :n])
    result['pb'][:, :n] = bp_lib.convToFloat16(struct,
                                               fdata['PA_LFR_SC_BP1_PB_' + freq][:, :n])
    # Compute n1 component of k-vector (normalized)
    result['nvec_v0'][:, :n] = bp_lib.convToFloat8(
        fdata['PA_LFR_SC_BP1_NVEC_V0_' + freq][:, :n])
    # Compute n2 component of k-vector (normalized)
    result['nvec_v1'][:, :n] = bp_lib.convToFloat8(
        fdata['PA_LFR_SC_BP1_NVEC_V1_' + freq][:, :n])

    # Store the 1b bit of the sign of n3 components k-vector in nvec_v2
    result['nvec_v2'][:, :n] = bp_lib.convToUint8(
        fdata['PA_LFR_SC_BP1_NVEC_' + freq][:, :n])

    result['ellip'][:, :n] = bp_lib.convToFloat4(
        fdata['PA_LFR_SC_BP1_ELLIP_' + freq][:, :n])

    result['dop'][:, :n] = bp_lib.convToFloat3(
        fdata['PA_LFR_SC_BP1_DOP_' + freq][:, :n])

    # Extract x-component of normalized Poynting flux...
    # sign (1 bit)...
    sx_sign = 1 - \
        (2 * bp_lib.Uint16ToUint8(fdata['PA_LFR_SC_BP1_SX_' +
                                        freq][:, :n], 15, 0x1).astype(np.float))

    # real value (14 bits)
    result['sx_rea'][:, :n] = sx_sign * bp_lib.convToFloat14(struct,
                                                             fdata['PA_LFR_SC_BP1_SX_' + freq][:, :n])

    # and arg (1 bit)
    result['sx_arg'][:, :n] = bp_lib.Uint16ToUint8(
        fdata['PA_LFR_SC_BP1_SX_' + freq][:, :n], 14, 0x1)

    # Extract velocity phase...

    # sign (1 bit)...
    vphi_sign = 1 - (2 * bp_lib.Uint16ToUint8(
        fdata['PA_LFR_SC_BP1_VPHI_' + freq][:, :n], 15, 0x1).astype(np.float))

    # real value (14 bits)
    result['vphi_rea'][:, :n] = vphi_sign * bp_lib.convToFloat14(struct,
                                                                 fdata['PA_LFR_SC_BP1_VPHI_' + freq][:, :n])

    # and arg (1 bit)
    result['vphi_arg'][:, :n] = bp_lib.Uint16ToUint8(
        fdata['PA_LFR_SC_BP1_VPHI_' + freq][:, :n], 14, 0x1)

    # Get TT2000 Epoch time
    result['epoch'] = Time().obt_to_utc(
        result['acquisition_time'], to_tt2000=True)

    # Get scet
    result['scet'] = Time.cuc_to_scet(result['acquisition_time'])

    return result

# Numpy array dtype for LFR BP2 L1 SBM1 data
# Name convention is lower case
BP2 = [
    ('epoch', 'uint64'),
    ('scet', 'float64'),
    ('acquisition_time', 'uint32', 2),
    ('synchro_flag', 'uint8'),
    ('quality_flag', 'uint8'),
    ('quality_bitmask', 'uint8'),
    ('bias_mode_mux_set', 'uint8'),
    ('bias_mode_hv_enabled', 'uint8'),
    ('bias_mode_bias1_enabled', 'uint8'),
    ('bias_mode_bias2_enabled', 'uint8'),
    ('bias_mode_bias3_enabled', 'uint8'),
    ('bias_on_off', 'uint8'),
    ('bw', 'uint8'),
    ('sp0', 'uint8'),
    ('sp1', 'uint8'),
    ('r0', 'uint8'),
    ('r1', 'uint8'),
    ('r2', 'uint8'),
    ('auto', 'float64', (22, 5)),
    ('cross_re', 'float32', (22, 10)),
    ('cross_im', 'float32', (22, 10)),
]


def bp2(f, task):
    # name of the packet
    name = 'TM_LFR_SCIENCE_SBM1_BP2_F0'
    logger.debug('Get data for ' + name)
    # get the sbm1 bp2 data
    if name in f['TM']:
        data = f['TM'][name]['source_data']
    else:
        return np.empty(
            0,
            dtype=BP2,
        )

    # Sort input packets by ascending acquisition times
    data, __ = sort_data(data, (
        data['PA_LFR_ACQUISITION_TIME'][:, 1],
        data['PA_LFR_ACQUISITION_TIME'][:, 0],
    ))

    # create an array for the data
    result = np.empty(
        data['PA_LFR_ACQUISITION_TIME'].shape[0],
        dtype=BP2,
    )
    fill_records(result)

    # get the acquisition time and other data
    result['acquisition_time'] = data['PA_LFR_ACQUISITION_TIME'][:, :2]
    result['synchro_flag'] = data['PA_LFR_ACQUISITION_TIME'][:, 2]
    result['bias_mode_mux_set'] = data['PA_BIA_MODE_MUX_SET']
    result['bias_mode_hv_enabled'] = data['PA_BIA_MODE_HV_ENABLED']
    result['bias_mode_bias1_enabled'] = data['PA_BIA_MODE_BIAS1_ENABLED']
    result['bias_mode_bias2_enabled'] = data['PA_BIA_MODE_BIAS2_ENABLED']
    result['bias_mode_bias3_enabled'] = data['PA_BIA_MODE_BIAS3_ENABLED']
    result['bias_on_off'] = data['PA_BIA_ON_OFF']
    result['bw'] = data['SY_LFR_BW']
    result['sp0'] = data['SY_LFR_SP0']
    result['sp1'] = data['SY_LFR_SP1']
    result['r0'] = data['SY_LFR_R0']
    result['r1'] = data['SY_LFR_R1']
    result['r2'] = data['SY_LFR_R2']

    auto = np.dstack(
        list(
            data['PA_LFR_SC_BP2_AUTO_A{0}_F0'.format(x)]
            for x in range(5)
        )
    )
    cross_re = np.dstack(
        list(
            data['PA_LFR_SC_BP2_CROSS_RE_{0}_F0'.format(x)]
            for x in range(10)
        )
    )
    cross_im = np.dstack(
        list(
            data['PA_LFR_SC_BP2_CROSS_IM_{0}_F0'.format(x)]
            for x in range(10)
        )
    )

    # Get number of packets and blocks
    shape = data['PA_LFR_SC_BP2_AUTO_A0_F0'].shape
    n = shape[1]

    # Convert AUTO, CROSS_RE an CROSS_IM (see convert_BP.py)
    struct = namedtuple('struct', ['nbitexp', 'nbitsig', 'rangesig_16', 'rangesig_14',
                                   'expmax', 'expmin'])
    bp_lib.init_struct(struct)

    result['auto'][:, :n, :] = bp_lib.convToAutoFloat(struct, auto)
    result['cross_re'][:, :n, :] = bp_lib.convToCrossFloat(cross_re)
    result['cross_im'][:, :n, :] = bp_lib.convToCrossFloat(cross_im)

    # WARNING, IF NAN or INF, put FILLVAL
    # TODO - Optimize this part
    result['auto'][np.isinf(result['auto'])] = -1e31
    result['auto'][np.isnan(result['auto'])] = -1e31
    result['cross_re'][np.isinf(result['cross_re'])] = -1e31
    result['cross_re'][np.isnan(result['cross_re'])] = -1e31
    result['cross_im'][np.isinf(result['cross_im'])] = -1e31
    result['cross_im'][np.isnan(result['cross_im'])] = -1e31

    # Get Epoch in nanoseconds
    result['epoch'] = Time().obt_to_utc(
        result['acquisition_time'], to_tt2000=True)

    # Get scet
    result['scet'] = Time.cuc_to_scet(result['acquisition_time'])

    return result

# Numpy array dtype for LFR CWF L1 SBM1 data
# Name convention is lower case
CWF = [
    ('epoch', 'uint64'),
    ('scet', 'float64'),
    ('acquisition_time', 'uint32', 2),
    ('synchro_flag', 'uint8'),
    ('quality_flag', 'uint8'),
    ('quality_bitmask', 'uint8'),
    ('bias_mode_mux_set', 'uint8'),
    ('bias_mode_hv_enabled', 'uint8'),
    ('bias_mode_bias1_enabled', 'uint8'),
    ('bias_mode_bias2_enabled', 'uint8'),
    ('bias_mode_bias3_enabled', 'uint8'),
    ('bias_on_off', 'uint8'),
    ('bw', 'uint8'),
    ('sp0', 'uint8'),
    ('sp1', 'uint8'),
    ('r0', 'uint8'),
    ('r1', 'uint8'),
    ('r2', 'uint8'),
    ('freq', 'uint8'),
    ('sampling_rate', 'float32'),
    ('v', 'int16'),
    ('e', 'int16', 2),
    ('b', 'int16', 3),
]


def cwf(f, task):
    # name of the packet
    name = 'TM_LFR_SCIENCE_SBM1_CWF_F1'
    logger.debug('Get data for ' + name)

    # get the sbm1 bp1 data
    if name in f['TM']:
        fdata = f['TM'][name]['source_data']
    else:
        return np.empty(
            0,
            dtype=CWF,
        )

    in_data = {
        name: fdata[name][...] for name in fdata.keys()
    }

    # get the data for cwf
    out_data = set_cwf(in_data, 'F1', CWF, True, task)

    # specific parameters
    out_data['freq'][:] = 1

    return out_data
